import logging

import click
from linkml_runtime.utils.formatutils import camelcase, underscore
from linkml_runtime.utils.schemaview import SchemaView

from linkml.utils.generator import Generator


def _map_range_to_dbml_type(range_name: str) -> str:
    """
    Map LinkML range types to DBML types.

    :param range_name: LinkML range name
    :return: Corresponding DBML type
    """
    type_mapping = {
        "string": "varchar",
        "integer": "int",
        "float": "float",
        "boolean": "boolean",
        "date": "date",
        "datetime": "datetime",
    }
    return type_mapping.get(range_name, "varchar")  # Default to varchar


class DBMLGenerator(Generator):
    """
    A generator for converting a LinkML schema into DBML (Database Markup Language).
    """

    generatorname = "dbmlgen"
    generatorversion = "0.2.0"
    valid_formats = ["dbml"]

    def __post_init__(self) -> None:
        super().__post_init__()
        self.logger = logging.getLogger(__name__)
        self.schemaview = SchemaView(self.schema)

    def serialize(self) -> str:
        """
        Generate DBML representation of the LinkML schema.

        :return: DBML as a string
        """
        dbml_lines = [
            "// DBML generated from LinkML schema\n",
            f"Project {{\n  name: '{self.schemaview.schema.name}'\n}}\n",
        ]

        for class_name, class_def in self.schemaview.all_classes().items():
            dbml_lines.append(self._generate_table(class_name, class_def))

        # Generate relationships if applicable
        relationships = self._generate_relationships()
        if relationships:
            dbml_lines.append(relationships)

        return "\n".join(dbml_lines)

    def _generate_table(self, class_name: str, class_def) -> str:
        """
        Generate the DBML for a single class (table).

        :param class_name: Name of the class
        :param class_def: ClassDefinition object
        :return: DBML representation of the class
        """
        dbml = [f"Table {camelcase(class_name)} {{"]

        for slot_name in self.schemaview.class_induced_slots(class_name):
            slot = self.schemaview.get_slot(slot_name.name)
            dbml.append(self._generate_column(slot))

        dbml.append("}\n")
        return "\n".join(dbml)

    def _generate_column(self, slot) -> str:
        """
        Generate the DBML for a single slot (column).

        :param slot: SlotDefinition object
        :return: DBML representation of the column
        """
        column_name = slot.name
        data_type = _map_range_to_dbml_type(slot.range or "string")
        constraints = []
        constraints_str = ""

        if slot.required:
            constraints.append("not null")
        if slot.identifier:
            constraints.append("primary key")

        if constraints:
            constraints_str = f"{', '.join(constraints)}"
            constraints_str = "[" + constraints_str + "]"

        return f"  {underscore(column_name)} {data_type} {constraints_str}".strip()

    def _generate_relationships(self) -> str:
        """
        Generate DBML relationships based on slot ranges referencing other classes.

        :return: DBML representation of relationships
        """
        relationships = []
        for class_name, class_def in self.schemaview.all_classes().items():
            for slot_name in self.schemaview.class_induced_slots(class_name):
                slot = self.schemaview.get_slot(slot_name.name)

                # Check if the slot references another class
                if slot.range in self.schemaview.all_classes():
                    # Find the identifier slot of the referenced class
                    identifier_slot_name = next(
                        (
                            slot_name.name
                            for slot_name in self.schemaview.class_induced_slots(slot.range)
                            if self.schemaview.get_slot(slot_name.name).identifier
                        ),
                        None,
                    )

                    if identifier_slot_name is None:
                        raise ValueError(f"Referenced class '{slot.range}' does not have an identifier slot.")

                    # Generate the DBML relationship
                    relationships.append(
                        f"Ref: {camelcase(class_name)}.{underscore(slot.name)} > "
                        f"{camelcase(slot.range)}.{underscore(identifier_slot_name)}"
                    )
        return "\n".join(relationships)


# CLI Definition
@click.command()
@click.option(
    "--schema",
    "-s",
    required=True,
    type=click.Path(exists=True, dir_okay=False, file_okay=True),
    help="Path to the LinkML schema YAML file",
)
@click.option(
    "--output",
    "-o",
    required=False,
    type=click.Path(dir_okay=False, writable=True),
    help="Path to save the generated DBML file. If not specified, DBML will be printed to stdout.",
)
def cli(schema, output):
    """
    CLI for LinkML to DBML generator.
    """
    generator = DBMLGenerator(schema)

    # Generate the DBML
    dbml_output = generator.serialize()

    # Save to file or print to stdout
    if output:
        with open(output, "w", encoding="utf-8") as f:
            f.write(dbml_output)
        click.echo(f"DBML has been saved to {output}")
    else:
        click.echo(dbml_output)


if __name__ == "__main__":
    cli()



from typing import TYPE_CHECKING

from linkml_runtime.linkml_model.meta import PermissibleValue, PermissibleValueText

if TYPE_CHECKING:
    from .dataframe_generator import DataframeGenerator


class EnumHandlerBase:
    def __init__(self, generator: "DataframeGenerator"):
        self.generator = generator

    def escape_permissible_value_text(self, pv_text):
        return pv_text.replace("'", "\\'").replace('"', '\\"')

    def extract_permissible_text(self, pv):
        if isinstance(pv, str) or isinstance(pv, PermissibleValueText):
            return self.escape_permissible_value_text(pv)
        elif isinstance(pv, PermissibleValue):
            return pv.text.code
        else:
            raise ValueError(f"Invalid permissible value in enum : {pv}")

    def get_enum_permissible_values(self, enum):
        return list(map(self.extract_permissible_text, enum.permissible_values or []))



import logging
import os
from dataclasses import dataclass
from typing import Optional

import click
from jinja2 import Template
from linkml_runtime.linkml_model.meta import (
    ClassDefinition,
    ClassDefinitionName,
    Element,
    SlotDefinition,
    SlotDefinitionName,
)
from linkml_runtime.utils.formatutils import camelcase, underscore
from linkml_runtime.utils.schemaview import SchemaView

from linkml._version import __version__
from linkml.generators.oocodegen import OOCodeGenerator
from linkml.utils.generator import shared_arguments

logger = logging.getLogger(__name__)


type_map = {
    "str": "string",
    "int": "number",
    "Bool": "boolean",
    "float": "number",
    "XSDDate": "date",
}

type_init_map = {
    "str": "''",
    "int": "0",
    "Bool": "false",
    "float": "0.0",
    "XSDDate": "new Date()",
}

default_template = """
{%- for c in view.all_classes().values() -%}
{%- set cref = gen.classref(c) -%}
{% if cref -%}
export type {{cref}} = string;
{% endif -%}
{%- endfor -%}

{% for e in enums.values() %}
{%- if e.description -%}
/**
* {{e.description}}
*/
{%- endif %}
export enum {{e.name}} {
    {% if not e.values -%}
    dummy = "dummy"
    {%- endif %}
    {%- for _, pv in e['values'].items() %}
    {% if pv.description -%}
    /** {{pv.description}} */
    {% endif -%}
    {{pv.label}} = "{{pv.value}}",
    {%- endfor %}
};
{% endfor %}

{% for c in view.all_classes().values() -%}
{%- if c.description -%}
/**
 * {{c.description}}
 */
{%- endif -%}
{% set parents = gen.parents(c) %}
export interface {{gen.name(c)}} {%- if parents %} extends {{parents|join(', ')}} {%- endif %} {
    {%- for sn in view.class_slots(c.name, direct=not(gen.include_induced_slots)) %}
    {% set s = view.induced_slot(sn, c.name) %}
    {%- if s.description -%}
    /** {{s.description}} */
    {% endif -%}
    {{gen.name(s)}}{%- if not s.required -%}?{%- endif -%}: {{gen.range(s)}},
    {%- endfor %}
}

{% if gen.gen_type_utils %}
export function is{{gen.name(c)}}(o: object): o is {{gen.name(c)}} {
    {%- set rcs = gen.required_slots(c) %}
    {%- set comp = "&&" if rcs else "||" %}
    {%- set cs = rcs if rcs else view.class_slots(c.name, direct=False) %}
    return (
        {%- for sn in cs %}
        '{{sn}}' in o {%- if not loop.last %} {{comp}}{% endif -%}
        {%- endfor %}
    )
}

export function to{{gen.name(c)}}(o: {{gen.name(c)}}): {{gen.name(c)}} {
    return {
        {%- for sn in view.class_slots(c.name, direct=False) %}
        {%- set s = view.induced_slot(sn, c.name) %}
        {{sn}}: o.{{sn}} ?? {{gen.init_range(s)}}{%- if not loop.last %},{%- endif -%}
        {%- endfor %}
    }
}
{% endif %}
{% endfor %}
"""


@dataclass
class TypescriptGenerator(OOCodeGenerator):
    """
    Generates typescript from a schema
    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.0.1"
    valid_formats = ["text"]
    uses_schemaloader = False

    # ObjectVars
    gen_type_utils: bool = False
    include_induced_slots: bool = False

    def serialize(self, output=None) -> str:
        """Serialize a schema to typescript string"""

        sv: SchemaView = self.schemaview
        enums = self.generate_enums(sv.all_enums())
        template_obj = Template(default_template)
        out_str = template_obj.render(
            gen=self,
            schema=self.schemaview.schema,
            view=self.schemaview,
            enums=enums,
        )
        if output is not None:
            with open(output, "w") as out:
                out.write(out_str)
        return out_str

    @staticmethod
    def name(element: Element) -> str:
        """
        Returns the name of the element in its canonical form

        :param element:
        :return:
        """
        alias = element.name
        if isinstance(element, SlotDefinition) and element.alias:
            alias = element.alias
        if type(element).class_name == "slot_definition":
            return underscore(alias)
        else:
            return camelcase(alias)

    def classref(self, cls: ClassDefinition) -> Optional[str]:
        """
        Returns the class name for the class that holds a reference (foreign key) to members of this class

        E.g. if a class Person has an identifier field called unique_id, then this will
        return PersonUniqueId

        :param cls:
        :return: ref name, None if no identifier
        """
        id_slot = self.get_identifier_or_key_slot(cls.name)
        if id_slot:
            return f"{self.name(cls)}{camelcase(id_slot.name)}"
        else:
            return None

    def get_identifier_or_key_slot(self, cn: ClassDefinitionName) -> Optional[SlotDefinition]:
        sv = self.schemaview
        id_slot = sv.get_identifier_slot(cn)
        if id_slot:
            return id_slot
        else:
            for s in sv.class_induced_slots(cn):
                if s.key:
                    return s
            return None

    def range(self, slot: SlotDefinition) -> str:
        sv = self.schemaview
        r = slot.range
        if r in sv.all_classes():
            rc = sv.get_class(r)
            rc_ref = self.classref(rc)
            rc_name = self.name(rc)
            id_slot = self.get_identifier_or_key_slot(r)
            if slot.multivalued:
                if not id_slot or slot.inlined:
                    if slot.inlined_as_list or not id_slot:
                        return f"{rc_name}[]"
                    else:
                        return f"{{[index: {rc_ref}]: {rc_name} }}"
                else:
                    return f"{rc_ref}[]"
            else:
                if not id_slot or slot.inlined:
                    return rc_name
                else:
                    return f"{rc_ref}"
        else:
            if r in sv.all_types():
                t = sv.get_type(r)
                tsrange = "string"
                if t.base and t.base in type_map:
                    tsrange = type_map[t.base]
                elif t.typeof and t.typeof in type_map:
                    tsrange = type_map[t.typeof]
                else:
                    logger.warning(f"Unknown type.base: {t.name}")
                if slot.multivalued:
                    tsrange = f"{tsrange}[]"
                return tsrange
            return "string"

    def init_range(self, slot: SlotDefinition) -> str:
        sv = self.schemaview
        r = slot.range
        if r in sv.all_classes():
            id_slot = self.get_identifier_or_key_slot(r)
            if slot.multivalued:
                if not id_slot or slot.inlined:
                    if slot.inlined_as_list or not id_slot:
                        return "[]"
                    else:
                        return "{}"
                else:
                    return "[]"
            else:
                if not id_slot or slot.inlined:
                    return "{}"
                else:
                    return "null"
        else:
            if r in sv.all_types():
                t = sv.get_type(r)
                if slot.multivalued:
                    return "[]"
                elif t.base and t.base in type_map:
                    return type_init_map[t.base]
                elif t.typeof and t.typeof in type_map:
                    return type_init_map[t.typeof]
                else:
                    logger.warning(f"Unknown type.base: {t.name}")
                return "null"
            return "null"

    @staticmethod
    def parents(cls: ClassDefinition) -> list[ClassDefinitionName]:
        if cls.is_a:
            parents = [cls.is_a]
        else:
            parents = []
        return [ClassDefinitionName(camelcase(p)) for p in parents + cls.mixins]

    def default_value_for_type(self, typ: str) -> str:
        pass

    def required_slots(self, cls: ClassDefinition) -> list[SlotDefinitionName]:
        return [s for s in self.schemaview.class_slots(cls.name) if self.schemaview.induced_slot(s, cls.name).required]


@shared_arguments(TypescriptGenerator)
@click.version_option(__version__, "-V", "--version")
@click.option("--gen-type-utils/", "-u", help="Generate Type checking utils", is_flag=True)
@click.option("--include-induced-slots/", help="Generate slots induced through inheritance", is_flag=True)
@click.option("--output", type=click.Path(dir_okay=False))
@click.command()
def cli(yamlfile, gen_type_utils=False, include_induced_slots=False, output=None, **args):
    """Generate typescript interfaces and types

    See https://github.com/linkml/linkml-runtime.js
    """
    gen = TypescriptGenerator(
        yamlfile, gen_type_utils=gen_type_utils, include_induced_slots=include_induced_slots, **args
    )
    serialized = gen.serialize(output=output)
    if output is None:
        print(serialized)


if __name__ == "__main__":
    cli()



# Auto generated from config.yaml by pythongen.py version: 0.0.1
# Generation date: 2025-09-30T13:24:19
# Schema: linter-config
#
# id: https://w3id.org/linkml/linter/config
# description: A datamodel describing the configuration file accepted by the linkml-lint command
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import re
from dataclasses import dataclass
from datetime import date, datetime, time
from typing import Any, ClassVar, Dict, List, Optional, Union

from jsonasobj2 import JsonObj, as_dict
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from linkml_runtime.utils.formatutils import camelcase, sfx, underscore
from linkml_runtime.utils.metamodelcore import bnode, empty_dict, empty_list
from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_float, extended_int, extended_str
from rdflib import Namespace, URIRef

from linkml_runtime.linkml_model.types import Boolean, String
from linkml_runtime.utils.metamodelcore import Bool

metamodel_version = "1.7.0"
version = None

# Namespaces
LINKML = CurieNamespace("linkml", "https://w3id.org/linkml/")
LINTCFG = CurieNamespace("lintcfg", "https://w3id.org/linkml/linter/config")
DEFAULT_ = LINTCFG


# Types

# Class references


@dataclass(repr=False)
class Config(YAMLRoot):
    """
    This is the top-level representation of a LinkML linter configuration file. It allows defining a set of rules
    while also optionally extending a predefined set of rules.
    """

    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = LINTCFG["Config"]
    class_class_curie: ClassVar[str] = "lintcfg:Config"
    class_name: ClassVar[str] = "Config"
    class_model_uri: ClassVar[URIRef] = LINTCFG.Config

    extends: Optional[Union[str, "ExtendableConfigs"]] = None
    rules: Optional[Union[dict, "Rules"]] = None

    def __post_init__(self, *_: str, **kwargs: Any):
        if self.extends is not None and not isinstance(self.extends, ExtendableConfigs):
            self.extends = ExtendableConfigs(self.extends)

        if self.rules is not None and not isinstance(self.rules, Rules):
            self.rules = Rules(**as_dict(self.rules))

        super().__post_init__(**kwargs)


@dataclass(repr=False)
class Rules(YAMLRoot):
    """
    Each attribute of this class represents a rule that can be enabled and possibly configured by a configuration file.
    """

    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = LINTCFG["Rules"]
    class_class_curie: ClassVar[str] = "lintcfg:Rules"
    class_name: ClassVar[str] = "Rules"
    class_model_uri: ClassVar[URIRef] = LINTCFG.Rules

    canonical_prefixes: Optional[Union[dict, "CanonicalPrefixesConfig"]] = None
    no_empty_title: Optional[Union[dict, "NoEmptyTitleConfig"]] = None
    no_invalid_slot_usage: Optional[Union[dict, "RuleConfig"]] = None
    no_undeclared_slots: Optional[Union[dict, "RuleConfig"]] = None
    no_undeclared_ranges: Optional[Union[dict, "RuleConfig"]] = None
    no_xsd_int_type: Optional[Union[dict, "RuleConfig"]] = None
    one_identifier_per_class: Optional[Union[dict, "RuleConfig"]] = None
    one_key_per_class: Optional[Union[dict, "RuleConfig"]] = None
    permissible_values_format: Optional[Union[dict, "PermissibleValuesFormatRuleConfig"]] = None
    recommended: Optional[Union[dict, "RecommendedRuleConfig"]] = None
    root_type_checks: Optional[Union[dict, "RuleConfig"]] = None
    standard_naming: Optional[Union[dict, "StandardNamingConfig"]] = None
    tree_root_class: Optional[Union[dict, "TreeRootClassRuleConfig"]] = None

    def __post_init__(self, *_: str, **kwargs: Any):
        if self.canonical_prefixes is not None and not isinstance(self.canonical_prefixes, CanonicalPrefixesConfig):
            self.canonical_prefixes = CanonicalPrefixesConfig(**as_dict(self.canonical_prefixes))

        if self.no_empty_title is not None and not isinstance(self.no_empty_title, NoEmptyTitleConfig):
            self.no_empty_title = NoEmptyTitleConfig(**as_dict(self.no_empty_title))

        if self.no_invalid_slot_usage is not None and not isinstance(self.no_invalid_slot_usage, RuleConfig):
            self.no_invalid_slot_usage = RuleConfig(**as_dict(self.no_invalid_slot_usage))

        if self.no_undeclared_slots is not None and not isinstance(self.no_undeclared_slots, RuleConfig):
            self.no_undeclared_slots = RuleConfig(**as_dict(self.no_undeclared_slots))

        if self.no_undeclared_ranges is not None and not isinstance(self.no_undeclared_ranges, RuleConfig):
            self.no_undeclared_ranges = RuleConfig(**as_dict(self.no_undeclared_ranges))

        if self.no_xsd_int_type is not None and not isinstance(self.no_xsd_int_type, RuleConfig):
            self.no_xsd_int_type = RuleConfig(**as_dict(self.no_xsd_int_type))

        if self.one_identifier_per_class is not None and not isinstance(self.one_identifier_per_class, RuleConfig):
            self.one_identifier_per_class = RuleConfig(**as_dict(self.one_identifier_per_class))

        if self.one_key_per_class is not None and not isinstance(self.one_key_per_class, RuleConfig):
            self.one_key_per_class = RuleConfig(**as_dict(self.one_key_per_class))

        if self.permissible_values_format is not None and not isinstance(
            self.permissible_values_format, PermissibleValuesFormatRuleConfig
        ):
            self.permissible_values_format = PermissibleValuesFormatRuleConfig(
                **as_dict(self.permissible_values_format)
            )

        if self.recommended is not None and not isinstance(self.recommended, RecommendedRuleConfig):
            self.recommended = RecommendedRuleConfig(**as_dict(self.recommended))

        if self.root_type_checks is not None and not isinstance(self.root_type_checks, RuleConfig):
            self.root_type_checks = RuleConfig(**as_dict(self.root_type_checks))

        if self.standard_naming is not None and not isinstance(self.standard_naming, StandardNamingConfig):
            self.standard_naming = StandardNamingConfig(**as_dict(self.standard_naming))

        if self.tree_root_class is not None and not isinstance(self.tree_root_class, TreeRootClassRuleConfig):
            self.tree_root_class = TreeRootClassRuleConfig(**as_dict(self.tree_root_class))

        super().__post_init__(**kwargs)


@dataclass(repr=False)
class RuleConfig(YAMLRoot):
    """
    This is the base class for linter rules. It contains configuration options that are common to all rules.
    """

    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = LINTCFG["RuleConfig"]
    class_class_curie: ClassVar[str] = "lintcfg:RuleConfig"
    class_name: ClassVar[str] = "RuleConfig"
    class_model_uri: ClassVar[URIRef] = LINTCFG.RuleConfig

    level: Union[str, "RuleLevel"] = None

    def __post_init__(self, *_: str, **kwargs: Any):
        if self._is_empty(self.level):
            self.MissingRequiredField("level")
        if not isinstance(self.level, RuleLevel):
            self.level = RuleLevel(self.level)

        super().__post_init__(**kwargs)


@dataclass(repr=False)
class PermissibleValuesFormatRuleConfig(RuleConfig):
    """
    Additional configuration options for the `permissible_values_format` rule
    """

    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = LINTCFG["PermissibleValuesFormatRuleConfig"]
    class_class_curie: ClassVar[str] = "lintcfg:PermissibleValuesFormatRuleConfig"
    class_name: ClassVar[str] = "PermissibleValuesFormatRuleConfig"
    class_model_uri: ClassVar[URIRef] = LINTCFG.PermissibleValuesFormatRuleConfig

    level: Union[str, "RuleLevel"] = None
    format: Optional[str] = None

    def __post_init__(self, *_: str, **kwargs: Any):
        if self.format is not None and not isinstance(self.format, str):
            self.format = str(self.format)

        super().__post_init__(**kwargs)


@dataclass(repr=False)
class TreeRootClassRuleConfig(RuleConfig):
    """
    Additional configuration options for the `tree_root_class` rule
    """

    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = LINTCFG["TreeRootClassRuleConfig"]
    class_class_curie: ClassVar[str] = "lintcfg:TreeRootClassRuleConfig"
    class_name: ClassVar[str] = "TreeRootClassRuleConfig"
    class_model_uri: ClassVar[URIRef] = LINTCFG.TreeRootClassRuleConfig

    level: Union[str, "RuleLevel"] = None
    root_class_name: Optional[str] = None
    validate_existing_class_name: Optional[Union[bool, Bool]] = None

    def __post_init__(self, *_: str, **kwargs: Any):
        if self.root_class_name is not None and not isinstance(self.root_class_name, str):
            self.root_class_name = str(self.root_class_name)

        if self.validate_existing_class_name is not None and not isinstance(self.validate_existing_class_name, Bool):
            self.validate_existing_class_name = Bool(self.validate_existing_class_name)

        super().__post_init__(**kwargs)


@dataclass(repr=False)
class RecommendedRuleConfig(RuleConfig):
    """
    Additional configuration options for the `recommended` rule
    """

    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = LINTCFG["RecommendedRuleConfig"]
    class_class_curie: ClassVar[str] = "lintcfg:RecommendedRuleConfig"
    class_name: ClassVar[str] = "RecommendedRuleConfig"
    class_model_uri: ClassVar[URIRef] = LINTCFG.RecommendedRuleConfig

    level: Union[str, "RuleLevel"] = None
    include: Optional[Union[str, list[str]]] = empty_list()
    exclude: Optional[Union[str, list[str]]] = empty_list()
    exclude_type: Optional[
        Union[Union[str, "MetamodelElementTypeEnum"], list[Union[str, "MetamodelElementTypeEnum"]]]
    ] = empty_list()

    def __post_init__(self, *_: str, **kwargs: Any):
        if not isinstance(self.include, list):
            self.include = [self.include] if self.include is not None else []
        self.include = [v if isinstance(v, str) else str(v) for v in self.include]

        if not isinstance(self.exclude, list):
            self.exclude = [self.exclude] if self.exclude is not None else []
        self.exclude = [v if isinstance(v, str) else str(v) for v in self.exclude]

        if not isinstance(self.exclude_type, list):
            self.exclude_type = [self.exclude_type] if self.exclude_type is not None else []
        self.exclude_type = [
            v if isinstance(v, MetamodelElementTypeEnum) else MetamodelElementTypeEnum(v) for v in self.exclude_type
        ]

        super().__post_init__(**kwargs)


@dataclass(repr=False)
class StandardNamingConfig(RuleConfig):
    """
    Additional configuration options for the `standard_naming` rule
    """

    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = LINTCFG["StandardNamingConfig"]
    class_class_curie: ClassVar[str] = "lintcfg:StandardNamingConfig"
    class_name: ClassVar[str] = "StandardNamingConfig"
    class_model_uri: ClassVar[URIRef] = LINTCFG.StandardNamingConfig

    level: Union[str, "RuleLevel"] = None
    permissible_values_upper_case: Optional[Union[bool, Bool]] = None
    exclude_type: Optional[
        Union[Union[str, "MetamodelElementTypeEnum"], list[Union[str, "MetamodelElementTypeEnum"]]]
    ] = empty_list()
    exclude: Optional[Union[str, list[str]]] = empty_list()
    class_pattern: Optional[str] = None
    slot_pattern: Optional[str] = None

    def __post_init__(self, *_: str, **kwargs: Any):
        if self.permissible_values_upper_case is not None and not isinstance(self.permissible_values_upper_case, Bool):
            self.permissible_values_upper_case = Bool(self.permissible_values_upper_case)

        if not isinstance(self.exclude_type, list):
            self.exclude_type = [self.exclude_type] if self.exclude_type is not None else []
        self.exclude_type = [
            v if isinstance(v, MetamodelElementTypeEnum) else MetamodelElementTypeEnum(v) for v in self.exclude_type
        ]

        if not isinstance(self.exclude, list):
            self.exclude = [self.exclude] if self.exclude is not None else []
        self.exclude = [v if isinstance(v, str) else str(v) for v in self.exclude]

        if self.class_pattern is not None and not isinstance(self.class_pattern, str):
            self.class_pattern = str(self.class_pattern)

        if self.slot_pattern is not None and not isinstance(self.slot_pattern, str):
            self.slot_pattern = str(self.slot_pattern)

        super().__post_init__(**kwargs)


@dataclass(repr=False)
class CanonicalPrefixesConfig(RuleConfig):
    """
    Additional configuration options for the canonical_prefixes rule
    """

    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = LINTCFG["CanonicalPrefixesConfig"]
    class_class_curie: ClassVar[str] = "lintcfg:CanonicalPrefixesConfig"
    class_name: ClassVar[str] = "CanonicalPrefixesConfig"
    class_model_uri: ClassVar[URIRef] = LINTCFG.CanonicalPrefixesConfig

    level: Union[str, "RuleLevel"] = None
    prefixmaps_contexts: Optional[Union[str, list[str]]] = empty_list()

    def __post_init__(self, *_: str, **kwargs: Any):
        if not isinstance(self.prefixmaps_contexts, list):
            self.prefixmaps_contexts = [self.prefixmaps_contexts] if self.prefixmaps_contexts is not None else []
        self.prefixmaps_contexts = [v if isinstance(v, str) else str(v) for v in self.prefixmaps_contexts]

        super().__post_init__(**kwargs)


@dataclass(repr=False)
class NoEmptyTitleConfig(RuleConfig):
    """
    Additional configuration options for the no_empty_title rule
    """

    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = LINTCFG["NoEmptyTitleConfig"]
    class_class_curie: ClassVar[str] = "lintcfg:NoEmptyTitleConfig"
    class_name: ClassVar[str] = "NoEmptyTitleConfig"
    class_model_uri: ClassVar[URIRef] = LINTCFG.NoEmptyTitleConfig

    level: Union[str, "RuleLevel"] = None
    exclude_type: Optional[
        Union[Union[str, "MetamodelElementTypeEnum"], list[Union[str, "MetamodelElementTypeEnum"]]]
    ] = empty_list()

    def __post_init__(self, *_: str, **kwargs: Any):
        if not isinstance(self.exclude_type, list):
            self.exclude_type = [self.exclude_type] if self.exclude_type is not None else []
        self.exclude_type = [
            v if isinstance(v, MetamodelElementTypeEnum) else MetamodelElementTypeEnum(v) for v in self.exclude_type
        ]

        super().__post_init__(**kwargs)


# Enumerations
class ExtendableConfigs(EnumDefinitionImpl):
    """
    The permissible values for the `extends` field of a config file
    """

    recommended = PermissibleValue(text="recommended", description="Extend the recommended rule set")

    _defn = EnumDefinition(
        name="ExtendableConfigs",
        description="The permissible values for the `extends` field of a config file",
    )


class RuleLevel(EnumDefinitionImpl):
    """
    The permissible values for the `level` option of all rules
    """

    disabled = PermissibleValue(text="disabled", description="The rule will not be checked")
    warning = PermissibleValue(
        text="warning", description="A violation of a rule at this level is a minor issue that should be fixed"
    )
    error = PermissibleValue(
        text="error", description="A violation of a rule at this level is a major issue that must be fixed"
    )

    _defn = EnumDefinition(
        name="RuleLevel",
        description="The permissible values for the `level` option of all rules",
    )


class MetamodelElementTypeEnum(EnumDefinitionImpl):
    """
    The permissible values for the exclude_type slot
    """

    class_definition = PermissibleValue(text="class_definition", meaning=LINKML["ClassDefinition"])
    enum_definition = PermissibleValue(text="enum_definition", meaning=LINKML["EnumDefinition"])
    permissible_value = PermissibleValue(text="permissible_value", meaning=LINKML["PermissibleValue"])
    slot_definition = PermissibleValue(text="slot_definition", meaning=LINKML["SlotDefinition"])

    _defn = EnumDefinition(
        name="MetamodelElementTypeEnum",
        description="The permissible values for the exclude_type slot",
    )


# Slots
class slots:
    pass


slots.config__extends = Slot(
    uri=LINTCFG.extends,
    name="config__extends",
    curie=LINTCFG.curie("extends"),
    model_uri=LINTCFG.config__extends,
    domain=None,
    range=Optional[Union[str, "ExtendableConfigs"]],
)

slots.config__rules = Slot(
    uri=LINTCFG.rules,
    name="config__rules",
    curie=LINTCFG.curie("rules"),
    model_uri=LINTCFG.config__rules,
    domain=None,
    range=Optional[Union[dict, Rules]],
)

slots.rules__canonical_prefixes = Slot(
    uri=LINTCFG.canonical_prefixes,
    name="rules__canonical_prefixes",
    curie=LINTCFG.curie("canonical_prefixes"),
    model_uri=LINTCFG.rules__canonical_prefixes,
    domain=None,
    range=Optional[Union[dict, CanonicalPrefixesConfig]],
)

slots.rules__no_empty_title = Slot(
    uri=LINTCFG.no_empty_title,
    name="rules__no_empty_title",
    curie=LINTCFG.curie("no_empty_title"),
    model_uri=LINTCFG.rules__no_empty_title,
    domain=None,
    range=Optional[Union[dict, NoEmptyTitleConfig]],
)

slots.rules__no_invalid_slot_usage = Slot(
    uri=LINTCFG.no_invalid_slot_usage,
    name="rules__no_invalid_slot_usage",
    curie=LINTCFG.curie("no_invalid_slot_usage"),
    model_uri=LINTCFG.rules__no_invalid_slot_usage,
    domain=None,
    range=Optional[Union[dict, RuleConfig]],
)

slots.rules__no_undeclared_slots = Slot(
    uri=LINTCFG.no_undeclared_slots,
    name="rules__no_undeclared_slots",
    curie=LINTCFG.curie("no_undeclared_slots"),
    model_uri=LINTCFG.rules__no_undeclared_slots,
    domain=None,
    range=Optional[Union[dict, RuleConfig]],
)

slots.rules__no_undeclared_ranges = Slot(
    uri=LINTCFG.no_undeclared_ranges,
    name="rules__no_undeclared_ranges",
    curie=LINTCFG.curie("no_undeclared_ranges"),
    model_uri=LINTCFG.rules__no_undeclared_ranges,
    domain=None,
    range=Optional[Union[dict, RuleConfig]],
)

slots.rules__no_xsd_int_type = Slot(
    uri=LINTCFG.no_xsd_int_type,
    name="rules__no_xsd_int_type",
    curie=LINTCFG.curie("no_xsd_int_type"),
    model_uri=LINTCFG.rules__no_xsd_int_type,
    domain=None,
    range=Optional[Union[dict, RuleConfig]],
)

slots.rules__one_identifier_per_class = Slot(
    uri=LINTCFG.one_identifier_per_class,
    name="rules__one_identifier_per_class",
    curie=LINTCFG.curie("one_identifier_per_class"),
    model_uri=LINTCFG.rules__one_identifier_per_class,
    domain=None,
    range=Optional[Union[dict, RuleConfig]],
)

slots.rules__one_key_per_class = Slot(
    uri=LINTCFG.one_key_per_class,
    name="rules__one_key_per_class",
    curie=LINTCFG.curie("one_key_per_class"),
    model_uri=LINTCFG.rules__one_key_per_class,
    domain=None,
    range=Optional[Union[dict, RuleConfig]],
)

slots.rules__permissible_values_format = Slot(
    uri=LINTCFG.permissible_values_format,
    name="rules__permissible_values_format",
    curie=LINTCFG.curie("permissible_values_format"),
    model_uri=LINTCFG.rules__permissible_values_format,
    domain=None,
    range=Optional[Union[dict, PermissibleValuesFormatRuleConfig]],
)

slots.rules__recommended = Slot(
    uri=LINTCFG.recommended,
    name="rules__recommended",
    curie=LINTCFG.curie("recommended"),
    model_uri=LINTCFG.rules__recommended,
    domain=None,
    range=Optional[Union[dict, RecommendedRuleConfig]],
)

slots.rules__root_type_checks = Slot(
    uri=LINTCFG.root_type_checks,
    name="rules__root_type_checks",
    curie=LINTCFG.curie("root_type_checks"),
    model_uri=LINTCFG.rules__root_type_checks,
    domain=None,
    range=Optional[Union[dict, RuleConfig]],
)

slots.rules__standard_naming = Slot(
    uri=LINTCFG.standard_naming,
    name="rules__standard_naming",
    curie=LINTCFG.curie("standard_naming"),
    model_uri=LINTCFG.rules__standard_naming,
    domain=None,
    range=Optional[Union[dict, StandardNamingConfig]],
)

slots.rules__tree_root_class = Slot(
    uri=LINTCFG.tree_root_class,
    name="rules__tree_root_class",
    curie=LINTCFG.curie("tree_root_class"),
    model_uri=LINTCFG.rules__tree_root_class,
    domain=None,
    range=Optional[Union[dict, TreeRootClassRuleConfig]],
)

slots.ruleConfig__level = Slot(
    uri=LINTCFG.level,
    name="ruleConfig__level",
    curie=LINTCFG.curie("level"),
    model_uri=LINTCFG.ruleConfig__level,
    domain=None,
    range=Union[str, "RuleLevel"],
)

slots.permissibleValuesFormatRuleConfig__format = Slot(
    uri=LINTCFG.format,
    name="permissibleValuesFormatRuleConfig__format",
    curie=LINTCFG.curie("format"),
    model_uri=LINTCFG.permissibleValuesFormatRuleConfig__format,
    domain=None,
    range=Optional[str],
)

slots.treeRootClassRuleConfig__root_class_name = Slot(
    uri=LINTCFG.root_class_name,
    name="treeRootClassRuleConfig__root_class_name",
    curie=LINTCFG.curie("root_class_name"),
    model_uri=LINTCFG.treeRootClassRuleConfig__root_class_name,
    domain=None,
    range=Optional[str],
)

slots.treeRootClassRuleConfig__validate_existing_class_name = Slot(
    uri=LINTCFG.validate_existing_class_name,
    name="treeRootClassRuleConfig__validate_existing_class_name",
    curie=LINTCFG.curie("validate_existing_class_name"),
    model_uri=LINTCFG.treeRootClassRuleConfig__validate_existing_class_name,
    domain=None,
    range=Optional[Union[bool, Bool]],
)

slots.recommendedRuleConfig__include = Slot(
    uri=LINTCFG.include,
    name="recommendedRuleConfig__include",
    curie=LINTCFG.curie("include"),
    model_uri=LINTCFG.recommendedRuleConfig__include,
    domain=None,
    range=Optional[Union[str, list[str]]],
)

slots.recommendedRuleConfig__exclude = Slot(
    uri=LINTCFG.exclude,
    name="recommendedRuleConfig__exclude",
    curie=LINTCFG.curie("exclude"),
    model_uri=LINTCFG.recommendedRuleConfig__exclude,
    domain=None,
    range=Optional[Union[str, list[str]]],
)

slots.recommendedRuleConfig__exclude_type = Slot(
    uri=LINTCFG.exclude_type,
    name="recommendedRuleConfig__exclude_type",
    curie=LINTCFG.curie("exclude_type"),
    model_uri=LINTCFG.recommendedRuleConfig__exclude_type,
    domain=None,
    range=Optional[Union[Union[str, "MetamodelElementTypeEnum"], list[Union[str, "MetamodelElementTypeEnum"]]]],
)

slots.standardNamingConfig__permissible_values_upper_case = Slot(
    uri=LINTCFG.permissible_values_upper_case,
    name="standardNamingConfig__permissible_values_upper_case",
    curie=LINTCFG.curie("permissible_values_upper_case"),
    model_uri=LINTCFG.standardNamingConfig__permissible_values_upper_case,
    domain=None,
    range=Optional[Union[bool, Bool]],
)

slots.standardNamingConfig__exclude_type = Slot(
    uri=LINTCFG.exclude_type,
    name="standardNamingConfig__exclude_type",
    curie=LINTCFG.curie("exclude_type"),
    model_uri=LINTCFG.standardNamingConfig__exclude_type,
    domain=None,
    range=Optional[Union[Union[str, "MetamodelElementTypeEnum"], list[Union[str, "MetamodelElementTypeEnum"]]]],
)

slots.standardNamingConfig__exclude = Slot(
    uri=LINTCFG.exclude,
    name="standardNamingConfig__exclude",
    curie=LINTCFG.curie("exclude"),
    model_uri=LINTCFG.standardNamingConfig__exclude,
    domain=None,
    range=Optional[Union[str, list[str]]],
)

slots.standardNamingConfig__class_pattern = Slot(
    uri=LINTCFG.class_pattern,
    name="standardNamingConfig__class_pattern",
    curie=LINTCFG.curie("class_pattern"),
    model_uri=LINTCFG.standardNamingConfig__class_pattern,
    domain=None,
    range=Optional[str],
)

slots.standardNamingConfig__slot_pattern = Slot(
    uri=LINTCFG.slot_pattern,
    name="standardNamingConfig__slot_pattern",
    curie=LINTCFG.curie("slot_pattern"),
    model_uri=LINTCFG.standardNamingConfig__slot_pattern,
    domain=None,
    range=Optional[str],
)

slots.canonicalPrefixesConfig__prefixmaps_contexts = Slot(
    uri=LINTCFG.prefixmaps_contexts,
    name="canonicalPrefixesConfig__prefixmaps_contexts",
    curie=LINTCFG.curie("prefixmaps_contexts"),
    model_uri=LINTCFG.canonicalPrefixesConfig__prefixmaps_contexts,
    domain=None,
    range=Optional[Union[str, list[str]]],
)

slots.noEmptyTitleConfig__exclude_type = Slot(
    uri=LINTCFG.exclude_type,
    name="noEmptyTitleConfig__exclude_type",
    curie=LINTCFG.curie("exclude_type"),
    model_uri=LINTCFG.noEmptyTitleConfig__exclude_type,
    domain=None,
    range=Optional[Union[Union[str, "MetamodelElementTypeEnum"], list[Union[str, "MetamodelElementTypeEnum"]]]],
)



from .json_formatter import JsonFormatter
from .markdown_formatter import MarkdownFormatter
from .terminal_formatter import TerminalFormatter
from .tsv_formatter import TsvFormatter

__all__ = ["JsonFormatter", "MarkdownFormatter", "TerminalFormatter", "TsvFormatter"]



import logging
from copy import copy
from dataclasses import dataclass, field
from typing import Optional

from linkml_runtime.linkml_model import (
    Annotation,
    ClassDefinition,
    ClassDefinitionName,
    Definition,
    Prefix,
    SchemaDefinition,
    SlotDefinition,
)
from linkml_runtime.linkml_model.meta import UniqueKey
from linkml_runtime.utils.schemaview import SchemaView, SlotDefinitionName
from sqlalchemy import Enum

logger = logging.getLogger(__name__)


class RelationalAnnotations(Enum):
    PRIMARY_KEY = "primary_key"
    FOREIGN_KEY = "foreign_key"


class ForeignKeyPolicy(Enum):
    ALL_REFS_ARE_FKS = "all_refs_are_fks"
    INJECT_FK_FOR_NESTED = "inject_fk_for_nested"
    INJECT_FK_FOR_ALL_REFS = "inject_fk_for_all_refs"
    NO_FOREIGN_KEYS = "no_foreign_keys"


@dataclass
class Link:
    """
    Foreign key reference
    """

    source_class: Optional[str]  # optional for top-level slots
    source_slot: str
    target_class: str
    target_slot: str = None


@dataclass
class RelationalMapping:
    """
    Mapping between slot in source model and target in relational model

    Example, with join table created:

        RelationalMapping(source_class='Person',
            source_slot='aliases',
            target_class='Person_aliases',
            target_slot='aliases',
            uses_join_table=True)
    """

    source_class: str = None
    source_slot: str = None
    mapping_type: str = None
    target_class: str = None
    target_slot: str = None  # /
    join_class: str = None  # /
    uses_join_table: bool = None  # /  ## True if extra join table is created


@dataclass
class OneToAnyMapping(RelationalMapping):
    """
    A one-to-one or one-to-many mapping from a source class+slot to
    a target class+slot
    """

    target_class: str = None
    target_slot: str = None
    multivalued: bool = False


@dataclass
class ManyToManyMapping(RelationalMapping):
    """
    A many-to-many relationship introduces a join class/table

    See:

    - https://docs.sqlalchemy.org/en/14/orm/basic_relationships.html#relationships-many-to-many
    - https://stackoverflow.com/questions/5756559/how-to-build-many-to-many-relations-using-sqlalchemy-a-good-example
    """

    join_class: str = None  # aka secondary
    target_class: str = None  # actual target
    mapping_type: str = "ManyToMany"


@dataclass
class MultivaluedScalar(RelationalMapping):
    """
    See: https://docs.sqlalchemy.org/en/14/orm/extensions/associationproxy.html
    """

    join_class: str = None
    target_slot: str = None
    mapping_type: str = "MultivaluedScalar"


def add_attribute(attributes: dict[str, SlotDefinition], tgt_slot: SlotDefinition) -> None:
    attributes[tgt_slot.name] = tgt_slot


def add_annotation(element: Definition, tag: str, value: str) -> None:
    ann = Annotation(tag, value)
    element.annotations[ann.tag] = ann


def get_primary_key_attributes(cls: ClassDefinition) -> list[SlotDefinitionName]:
    return [a.name for a in cls.attributes.values() if RelationalAnnotations.PRIMARY_KEY in a.annotations]


def get_foreign_key_map(cls: ClassDefinition) -> dict[SlotDefinitionName, str]:
    return {
        a.name: a.annotations[RelationalAnnotations.FOREIGN_KEY].value
        for a in cls.attributes.values()
        if RelationalAnnotations.FOREIGN_KEY in a.annotations
    }


@dataclass
class TransformationResult:
    """
    The result of a transformation is a target schema plus a collection of mappings
    """

    schema: SchemaDefinition
    mappings: list[RelationalMapping]


@dataclass
class RelationalModelTransformer:
    """
    Transforms the source schema into a relational schema
    """

    schemaview: SchemaView = None
    # dialect: str = 'sqlite'
    skip_tree_root: bool = False
    skip_abstract: bool = True
    skip_mixins: bool = True
    join_table_separator: str = "_"
    foreign_key_policy: ForeignKeyPolicy = field(default_factory=lambda: ForeignKeyPolicy.INJECT_FK_FOR_NESTED)

    def transform(self, tgt_schema_name: str = None, top_class: ClassDefinitionName = None) -> TransformationResult:
        """
        Transforms the source schema into a relational schema

        :param tgt_schema_name:
        :param top_class:
        :return:
        """
        join_sep = self.join_table_separator
        links = self.get_reference_map()
        source_sv = self.schemaview
        source_sv.merge_imports()
        source = source_sv.schema
        src_schema_name = source.name
        mappings = []
        if tgt_schema_name is None:
            tgt_schema_name = f"{src_schema_name}_relational"
        tgt_schema_id = f"{source.id}_relational"
        # TODO: recursively transform imports
        target = SchemaDefinition(
            id=tgt_schema_id,
            name=tgt_schema_name,
            default_range=source.default_range,
            prefixes=source.prefixes,
            imports=source.imports,
            # imports=['linkml:types'],
            from_schema=source.from_schema,
            source_file=source.source_file,
            types=source.types,
            subsets=source.subsets,
            enums=source.enums,
            annotations=source.annotations,
        )
        target.prefixes["rr"] = Prefix("rr", "http://www.w3.org/ns/r2rml#")

        # copy source -> target
        # roll-down all slots and create an attribute-only model
        for cn, c in source_sv.all_classes().items():
            c = ClassDefinition(
                name=cn,
                class_uri=source_sv.get_uri(c, expand=False),
                mixin=c.mixin,
                is_a=c.is_a,
                tree_root=c.tree_root,
                abstract=c.abstract,
                description=c.description,
                unique_keys=c.unique_keys,
                annotations=c.annotations,
            )
            for slot in source_sv.class_induced_slots(cn):
                tgt_slot = copy(slot)
                if slot.alias:
                    tgt_slot.name = slot.alias
                # TODO: attrs not indexed
                # tgt_slot.slot_uri = sv.get_uri(slot, expand=False)
                tgt_slot.is_a = None
                tgt_slot.mixins = []
                add_attribute(c.attributes, tgt_slot)
            # this is required in case an attribute inherits from a slot
            for sn in source_sv.all_slots(attributes=False):
                slot = source_sv.get_slot(sn)
            target.classes[c.name] = c

        target_sv = SchemaView(target)
        # create surrogate/autoincrement primary keys for any class (originally: that is referenced)
        # for link in links:
        for cn in target_sv.all_classes():
            pk = self.get_direct_identifier_attribute(target_sv, cn)
            if self.foreign_key_policy == ForeignKeyPolicy.NO_FOREIGN_KEYS:
                logger.info(f"Will not inject any PKs, and policy == {self.foreign_key_policy}")
            else:
                if pk is None:
                    pk = self.add_primary_key(cn, target_sv)
                    logger.info(f"Added primary key {cn}.{pk.name}")
                for link in links:
                    if link.target_class == cn:
                        link.target_slot = pk.name

        # TODO: separate out the logic into separate testable methods
        target_sv.set_modified()
        multivalued_slots_original = []
        # post-process target schema
        for cn, c in target_sv.all_classes().items():
            if self.foreign_key_policy == ForeignKeyPolicy.NO_FOREIGN_KEYS:
                continue
            pk_slot = self.get_direct_identifier_attribute(target_sv, cn)
            # if self.is_skip(c) and len(incoming_links) == 0:
            #    logger.info(f'Skipping class: {c.name}')
            #    del target.classes[cn]
            #    continue
            for src_slot in list(c.attributes.values()):
                slot = copy(src_slot)
                slot_range = slot.range
                slot_range_is_class = slot_range in target_sv.all_classes()
                is_shared = slot_range_is_class and (
                    slot.inlined or slot.inlined_as_list or "shared" in slot.annotations
                )
                if slot.multivalued:
                    multivalued_slots_original.append(slot.name)
                    slot.multivalued = False
                    slot_name = slot.name
                    sn_singular = slot.singular_name if slot.singular_name else slot.name
                    if pk_slot is None:
                        pk_slot = self.add_primary_key(c.name, target_sv)
                    backref_slot = SlotDefinition(
                        name=f"{c.name}_{pk_slot.name}",
                        description="Autocreated FK slot",
                        range=c.name,
                        slot_uri="rdf:subject",
                        # close_mappings=[pk_slot.slot_uri],
                        annotations=[
                            Annotation("backref", "true"),
                            Annotation("rdfs:subPropertyOf", "rdf:subject"),
                        ],
                    )
                    # if is_only_ref_to_range and slot_range_is_class:
                    if is_shared:
                        # ONE-TO-MANY
                        # e.g. if Person->Address, and only Person has Address,
                        # we can make Address.Person_id
                        backref_slot.inverse = slot_name
                        backref_class = target.classes[slot_range]
                        add_attribute(backref_class.attributes, backref_slot)
                        # In SQLA, corresponds to source_class.source_slot = relationship(target_class)
                        mappings.append(
                            OneToAnyMapping(
                                source_class=cn,
                                source_slot=src_slot.name,
                                target_class=backref_class.name,
                                target_slot=backref_slot.name,
                            )
                        )
                        backref_key_slots = [s for s in backref_class.attributes.values() if s.key]
                        if backref_key_slots:
                            if len(backref_key_slots) > 1:
                                raise ValueError(f"Multiple keys for {c.name}: {backref_key_slots}")
                            backref_key_slot = backref_key_slots[0]
                            unique_key_name = f"{c.name}_{backref_key_slot.name}"
                            backref_class.unique_keys[unique_key_name] = UniqueKey(
                                unique_key_name=unique_key_name,
                                unique_key_slots=[backref_slot.name, backref_key_slot.name],
                            )

                    else:
                        # MANY-TO-MANY
                        # create new linking table
                        linker_class = ClassDefinition(
                            name=f"{cn}{join_sep}{sn_singular}",
                            from_schema=target.id,
                            class_uri="rdf:Statement",
                            annotations=[
                                Annotation("linkml:derived_from", cn),
                                Annotation("dcterms:conformsTo", "linkml:JoinTable"),
                            ],
                            comments=[f"Linking class generated from {cn}.{slot_name}"],
                        )
                        slot.name = sn_singular
                        # On the linking table, it's inlined.
                        # This triggers that the slot.name gets appended with the pk column name on the target side
                        slot.inlined = True
                        add_attribute(linker_class.attributes, backref_slot)
                        add_attribute(linker_class.attributes, slot)
                        slot.slot_uri = "rdf:object"
                        target.classes[linker_class.name] = linker_class
                        if slot_range_is_class:
                            fwdann = Annotation("forwardref", "true")
                            slot.annotations[fwdann.tag] = fwdann
                            mappings.append(
                                ManyToManyMapping(
                                    source_class=cn,
                                    source_slot=src_slot.name,
                                    target_class=slot_range,
                                    # target_slot=backref_slot.name,
                                    join_class=linker_class.name,
                                    # target_slot=slot.name,
                                    # uses_join_table=True,
                                )
                            )
                        else:
                            mappings.append(
                                MultivaluedScalar(
                                    source_class=cn,
                                    source_slot=src_slot.name,
                                    target_slot=sn_singular,
                                    join_class=linker_class.name,
                                )
                            )
                    # we delete the slot from the set of attributes for the class,
                    # but leave it present as a 'dangling' slot, where it can
                    # be referenced for mapping purposes
                    target.slots[slot_name] = src_slot
                    src_slot.owner = None
                    del c.attributes[slot_name]
                    target_sv.set_modified()
            target.classes[c.name] = c

        # add PK and FK anns
        target_sv.set_modified()
        fk_policy = self.foreign_key_policy
        forward_map = {}
        for c in target.classes.values():
            if self.foreign_key_policy == ForeignKeyPolicy.NO_FOREIGN_KEYS:
                continue
            pk_slot = target_sv.get_identifier_slot(c.name)
            for a in list(c.attributes.values()):
                if pk_slot is None or a.name == pk_slot.name:
                    ann = Annotation("primary_key", "true")
                    a.annotations[ann.tag] = ann
                if a.required:
                    ann = Annotation("required", "true")
                    a.annotations[ann.tag] = ann
                if a.range in target.classes:
                    tc = target.classes[a.range]
                    # tc_pk_slot = target_sv.get_identifier_slot(tc.name)
                    tc_pk_slot = self.get_direct_identifier_attribute(target_sv, tc.name)
                    if tc_pk_slot is None:
                        raise ValueError(f"No PK for attribute {a.name} range {a.range}")
                    is_inlined = a.inlined or not source_sv.get_identifier_slot(tc.name)
                    if (fk_policy == ForeignKeyPolicy.INJECT_FK_FOR_NESTED and is_inlined and not a.multivalued) or (
                        fk_policy == ForeignKeyPolicy.INJECT_FK_FOR_ALL_REFS
                    ):
                        # if it is already an injected backref, no need to re-inject
                        if "backref" not in a.annotations:
                            del c.attributes[a.name]
                            original_name = a.name
                            if "forwardref" not in a.annotations:
                                add_annotation(a, "original_slot", original_name)
                            a.alias = f"{a.name}_{tc_pk_slot.name}"
                            a.name = a.alias
                            c.attributes[a.name] = a
                            forward_map[original_name] = a.name
                    ann = Annotation("foreign_key", f"{tc.name}.{tc_pk_slot.name}")
                    a.annotations[ann.tag] = ann
                    target_sv.set_modified()
            # Rewrite unique key constraints
            # - if a slot has a range of object, it may be renamed, e.g. person => person_id
            # - if a slot is multivalued then it is translated to backref and the UC must be dropped
            removed_ucs = []
            for uc_name, uc in c.unique_keys.items():
                if any(sn in multivalued_slots_original for sn in uc.unique_key_slots):
                    logger.warning(
                        f"Cannot represent uniqueness constraint {uc_name}. "
                        f"one of the slots {uc.unique_key_slots} is multivalued"
                    )
                    removed_ucs.append(uc_name)
                new_slot_names = [forward_map.get(sn, sn) for sn in uc.unique_key_slots]
                uc.unique_key_slots = new_slot_names
            for uc_name in removed_ucs:
                del c.unique_keys[uc_name]

        result = TransformationResult(target, mappings=mappings)
        return result

    @staticmethod
    def get_direct_identifier_attribute(sv: SchemaView, cn: ClassDefinitionName) -> Optional[SlotDefinition]:
        c = sv.get_class(cn)
        for a in c.attributes.values():
            if a.identifier:
                return a
            if a.key:
                return a
        return None

    def get_reference_map(self) -> list[Link]:
        """
        Extract all class-slot-range references

        :return: list of links
        """
        # TODO: move this to schemaview
        links = []
        sv = self.schemaview
        for cn, c in sv.all_classes().items():
            for slot in sv.class_induced_slots(cn):
                if slot.range in sv.all_classes():
                    links.append(
                        Link(
                            source_class=cn,
                            source_slot=slot.name,
                            target_class=slot.range,
                        )
                    )
        for sn, slot in sv.all_slots().items():
            if slot.range in sv.all_classes():
                links.append(
                    Link(
                        source_class=None,
                        source_slot=slot.name,
                        target_class=slot.range,
                    )
                )
        return links

    def is_skip(self, c: ClassDefinition) -> bool:
        return (
            (c.abstract and self.skip_abstract)
            or (c.mixin and self.skip_mixins)
            or (c.tree_root and self.skip_tree_root)
        )

    @staticmethod
    def add_primary_key(cn: str, sv: SchemaView) -> SlotDefinition:
        """
        Adds a surrogate/autoincrement primary key to a class

        :param cn:
        :param sv:
        :return:
        """
        c = sv.get_class(cn)
        candidate_names = ["id", "uid", "identifier", "pk"]
        valid_candidate_names = [n for n in candidate_names if n not in c.attributes]
        if not valid_candidate_names:
            raise ValueError(f"Cannot add primary key to class {cn}: no valid candidate names")
        pk = SlotDefinition(name=valid_candidate_names[0], identifier=True, range="integer")
        add_annotation(pk, "dcterms:conformsTo", "rr:BlankNode")
        add_annotation(pk, "autoincrement", "true")
        if pk.name in c.attributes:
            raise ValueError(
                f"Cannot inject primary key {pk.name} as a non-unique attribute with this name already exists in {cn}"
            )
        # add PK to start of attributes
        atts = copy(c.attributes)
        c.attributes.clear()  # See https://github.com/linkml/linkml/issues/370
        add_attribute(c.attributes, pk)  # add to start
        c.attributes.update(atts)
        sv.set_modified()
        return pk



# Auto generated from reporting.yaml by pythongen.py version: 0.9.0
# Generation date: 2021-09-01 20:03
# Schema: reporting
#
# id: https://w3id.org/linkml/reporting
# description: A datamodel for reports on data
# license: https://creativecommons.org/publicdomain/zero/1.0/

from dataclasses import dataclass
from typing import Any, ClassVar, Optional, Union

from jsonasobj2 import as_dict
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from linkml_runtime.utils.metamodelcore import NodeIdentifier, URIorCURIE, empty_list
from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.yamlutils import YAMLRoot
from rdflib import URIRef

metamodel_version = "1.7.0"


# Namespaces
LINKML = CurieNamespace("linkml", "https://w3id.org/linkml/")
OWL = CurieNamespace("owl", "http://www.w3.org/2002/07/owl#")
PAV = CurieNamespace("pav", "http://purl.org/pav/")
RDF = CurieNamespace("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#")
RDFS = CurieNamespace("rdfs", "http://www.w3.org/2000/01/rdf-schema#")
REPORTING = CurieNamespace("reporting", "https://w3id.org/linkml/report")
SCHEMA = CurieNamespace("schema", "http://schema.org/")
SKOS = CurieNamespace("skos", "http://www.w3.org/2004/02/skos/core#")
XSD = CurieNamespace("xsd", "http://www.w3.org/2001/XMLSchema#")
DEFAULT_ = REPORTING


# Types

# Class references


@dataclass
class Report(YAMLRoot):
    """
    A report object
    """

    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = REPORTING.Report
    class_class_curie: ClassVar[str] = "reporting:Report"
    class_name: ClassVar[str] = "report"
    class_model_uri: ClassVar[URIRef] = REPORTING.Report

    results: Optional[Union[Union[dict, "CheckResult"], list[Union[dict, "CheckResult"]]]] = empty_list()

    def __post_init__(self, *_: list[str], **kwargs: dict[str, Any]):
        if not isinstance(self.results, list):
            self.results = [self.results] if self.results is not None else []
        self.results = [v if isinstance(v, CheckResult) else CheckResult(**as_dict(v)) for v in self.results]

        super().__post_init__(**kwargs)


@dataclass
class CheckResult(YAMLRoot):
    """
    An individual check
    """

    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = REPORTING.CheckResult
    class_class_curie: ClassVar[str] = "reporting:CheckResult"
    class_name: ClassVar[str] = "check_result"
    class_model_uri: ClassVar[URIRef] = REPORTING.CheckResult

    type: Optional[Union[str, URIorCURIE]] = None
    subject: Optional[Union[str, NodeIdentifier]] = None
    instantiates: Optional[Union[str, NodeIdentifier]] = None
    predicate: Optional[Union[str, NodeIdentifier]] = None
    object: Optional[Union[str, NodeIdentifier]] = None
    object_str: Optional[str] = None
    source: Optional[Union[str, NodeIdentifier]] = None
    info: Optional[str] = None

    def __post_init__(self, *_: list[str], **kwargs: dict[str, Any]):
        if self.type is not None and not isinstance(self.type, URIorCURIE):
            self.type = URIorCURIE(self.type)

        if self.subject is not None and not isinstance(self.subject, NodeIdentifier):
            self.subject = NodeIdentifier(self.subject)

        if self.instantiates is not None and not isinstance(self.instantiates, NodeIdentifier):
            self.instantiates = NodeIdentifier(self.instantiates)

        if self.predicate is not None and not isinstance(self.predicate, NodeIdentifier):
            self.predicate = NodeIdentifier(self.predicate)

        if self.object is not None and not isinstance(self.object, NodeIdentifier):
            self.object = NodeIdentifier(self.object)

        if self.object_str is not None and not isinstance(self.object_str, str):
            self.object_str = str(self.object_str)

        if self.source is not None and not isinstance(self.source, NodeIdentifier):
            self.source = NodeIdentifier(self.source)

        if self.info is not None and not isinstance(self.info, str):
            self.info = str(self.info)

        super().__post_init__(**kwargs)


class Problem(CheckResult):
    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = REPORTING.Problem
    class_class_curie: ClassVar[str] = "reporting:Problem"
    class_name: ClassVar[str] = "problem"
    class_model_uri: ClassVar[URIRef] = REPORTING.Problem


class ProblemSlotUndeclared(Problem):
    """
    A problem in which an undeclared slot is used
    """

    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = REPORTING.ProblemSlotUndeclared
    class_class_curie: ClassVar[str] = "reporting:ProblemSlotUndeclared"
    class_name: ClassVar[str] = "problem_slot_undeclared"
    class_model_uri: ClassVar[URIRef] = REPORTING.ProblemSlotUndeclared


class ProblemSlotInapplicable(Problem):
    """
    A problem in which a slot is used in an instance of a class where the slot is not applicable for that class
    """

    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = REPORTING.ProblemSlotInapplicable
    class_class_curie: ClassVar[str] = "reporting:ProblemSlotInapplicable"
    class_name: ClassVar[str] = "problem_slot_inapplicable"
    class_model_uri: ClassVar[URIRef] = REPORTING.ProblemSlotInapplicable


class ProblemSlotMissing(Problem):
    """
    A problem in which an instance of a class has a required slot which is not filled in
    """

    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = REPORTING.ProblemSlotMissing
    class_class_curie: ClassVar[str] = "reporting:ProblemSlotMissing"
    class_name: ClassVar[str] = "problem_slot_missing"
    class_model_uri: ClassVar[URIRef] = REPORTING.ProblemSlotMissing


# Enumerations
class SeverityOptions(EnumDefinitionImpl):
    FATAL = PermissibleValue(text="FATAL")
    ERROR = PermissibleValue(text="ERROR")
    WARNING = PermissibleValue(text="WARNING")
    INFO = PermissibleValue(text="INFO")

    _defn = EnumDefinition(
        name="SeverityOptions",
    )


# Slots
class slots:
    pass


slots.type = Slot(
    uri=REPORTING.type,
    name="type",
    curie=REPORTING.curie("type"),
    model_uri=REPORTING.type,
    domain=None,
    range=Optional[Union[str, URIorCURIE]],
)

slots.subject = Slot(
    uri=REPORTING.subject,
    name="subject",
    curie=REPORTING.curie("subject"),
    model_uri=REPORTING.subject,
    domain=None,
    range=Optional[Union[str, NodeIdentifier]],
)

slots.instantiates = Slot(
    uri=REPORTING.instantiates,
    name="instantiates",
    curie=REPORTING.curie("instantiates"),
    model_uri=REPORTING.instantiates,
    domain=None,
    range=Optional[Union[str, NodeIdentifier]],
)

slots.predicate = Slot(
    uri=REPORTING.predicate,
    name="predicate",
    curie=REPORTING.curie("predicate"),
    model_uri=REPORTING.predicate,
    domain=None,
    range=Optional[Union[str, NodeIdentifier]],
)

slots.object = Slot(
    uri=REPORTING.object,
    name="object",
    curie=REPORTING.curie("object"),
    model_uri=REPORTING.object,
    domain=None,
    range=Optional[Union[str, NodeIdentifier]],
)

slots.object_str = Slot(
    uri=REPORTING.object_str,
    name="object_str",
    curie=REPORTING.curie("object_str"),
    model_uri=REPORTING.object_str,
    domain=None,
    range=Optional[str],
)

slots.source = Slot(
    uri=REPORTING.source,
    name="source",
    curie=REPORTING.curie("source"),
    model_uri=REPORTING.source,
    domain=None,
    range=Optional[Union[str, NodeIdentifier]],
)

slots.severity = Slot(
    uri=REPORTING.severity,
    name="severity",
    curie=REPORTING.curie("severity"),
    model_uri=REPORTING.severity,
    domain=None,
    range=Optional[Union[str, "SeverityOptions"]],
)

slots.info = Slot(
    uri=REPORTING.info,
    name="info",
    curie=REPORTING.curie("info"),
    model_uri=REPORTING.info,
    domain=None,
    range=Optional[str],
)

slots.report__results = Slot(
    uri=REPORTING.results,
    name="report__results",
    curie=REPORTING.curie("results"),
    model_uri=REPORTING.report__results,
    domain=None,
    range=Optional[Union[Union[dict, CheckResult], list[Union[dict, CheckResult]]]],
)



import logging
import re
from collections import defaultdict
from copy import copy
from dataclasses import dataclass
from typing import Any, Callable, Optional, Union

import click
import yaml
from linkml_runtime import SchemaView
from linkml_runtime.dumpers import json_dumper
from linkml_runtime.linkml_model import (
    ClassDefinition,
    ClassDefinitionName,
    EnumDefinition,
    SchemaDefinition,
    SlotDefinition,
    TypeDefinition,
)
from linkml_runtime.utils.formatutils import camelcase, underscore
from linkml_runtime.utils.yamlutils import YAMLRoot

logger = logging.getLogger(__name__)

pattern = re.compile(r"(?<!^)(?=[A-Z])")


def yaml_rewrite(obj: Any, replacements: dict[str, Any], include_keys=True) -> Any:
    if isinstance(obj, YAMLRoot):
        obj2 = copy(obj)
        for k, v in vars(obj).items():
            setattr(obj2, k, yaml_rewrite(v, replacements, include_keys))
        return obj2
    if isinstance(obj, dict):
        if include_keys and set(obj.keys()).intersection(replacements):
            obj = {replacements.get(k, k): v for k, v in obj.items()}
        return {k: yaml_rewrite(v, replacements, include_keys) for k, v in obj.items()}
    if isinstance(obj, list):
        return [yaml_rewrite(v, replacements, include_keys) for v in obj]
    if isinstance(obj, str) and obj in replacements:
        return replacements.get(obj)
    return obj


def uncamel(n: str) -> str:
    return pattern.sub(" ", n)


@dataclass
class SchemaFixer:
    """
    Multiple methods for adding additional information to schemas
    """

    history: list[str] = None

    def add_titles(self, schema: SchemaDefinition):
        """
        Add titles to all elements if not present.

        Title will be generated from the name, expanding CamelCase => Camel Case,
        and replacing underscores with spaces.

        :param schema: input schema, will be modified in place
        :return:
        """
        sv = SchemaView(schema)
        for e in sv.all_elements().values():
            if e.title is not None:
                continue
            title = e.name.replace("_", " ")
            title = uncamel(title).lower()
            e.title = title
            self._add_history(f"added title {title} to {e.name}")

    def add_container(
        self,
        schema: SchemaDefinition,
        class_name: str = "Container",
        force: bool = False,
        **kwargs,
    ) -> ClassDefinition:
        """
        Adds a container class

        Also wraps :ref:`add_index_slots`

        :param schema: input schema, will be modified in place
        :param class_name: name for container
        :param force: if true, force adding a container, even if present
        :return: container class
        """
        sv = SchemaView(schema)
        tree_roots = [c for c in sv.all_classes().values() if c.tree_root]
        if len(tree_roots) > 0:
            if force:
                logger.info("Forcing addition of containers")
            else:
                raise ValueError(f"Schema already has containers: {tree_roots}")
        container = ClassDefinition(class_name, tree_root=True)
        sv.add_class(container)
        self._add_history(f"added container {container.name}")
        self.add_index_slots(schema, container.name, **kwargs)
        return container

    def add_index_slots(
        self,
        schema: SchemaDefinition,
        container_name: ClassDefinitionName,
        inlined_as_list=False,
        must_have_identifier=False,
        slot_name_func: Callable = None,
        convert_camel_case=False,
    ) -> list[SlotDefinition]:
        """
        Adds index slots to a container pointing at all top-level classes

        :param schema: input schema, will be modified in place
        :param container_name:
        :param inlined_as_list:
        :param must_have_identifier:
        :param slot_name_func: function to determine the name of the slot from the class
        :return: new slots
        """
        sv = SchemaView(schema)
        container = sv.get_class(container_name)
        ranges = set()
        for cn in sv.all_classes():
            for s in sv.class_induced_slots(cn):
                ranges.add(s.range)
        top_level_classes = [c for c in sv.all_classes().values() if not c.tree_root and c.name not in ranges]
        if must_have_identifier:
            top_level_classes = [c for c in top_level_classes if sv.get_identifier_slot(c.name) is not None]
        index_slots = []
        for c in top_level_classes:
            has_identifier = sv.get_identifier_slot(c.name)
            if slot_name_func:
                sn = slot_name_func(c)
            else:
                cn = c.name
                if convert_camel_case:
                    cn = uncamel(cn).lower()
                cn = cn.replace(" ", "_")
                sn = f"{cn}_index"
            index_slot = SlotDefinition(
                sn,
                range=c.name,
                multivalued=True,
                inlined_as_list=not has_identifier or inlined_as_list,
            )
            index_slots.append(index_slot)
            schema.slots[index_slot.name] = index_slot
            container.slots.append(index_slot.name)
            self._add_history(f"Adding container slot: {index_slot.name}")
        return index_slots

    def attributes_to_slots(self, schema: SchemaDefinition, remove_redundant_slot_usage=True) -> None:
        """
        Convert all attributes to slots

        :param schema:
        :param remove_redundant_slot_usage:
        :return:
        """
        sv = SchemaView(schema)
        new_slots = []
        for c in sv.all_classes().values():
            for a in c.attributes:
                new_slots.append(a)
                self.merge_slot_usage(sv, c, a)
                del c.attributes[a.name]
        for slot in new_slots:
            if slot.name in sv.all_slots():
                raise ValueError(f"Duplicate slot {slot.name}")
            sv.add_slot(slot)
            self._add_history(f"Adding slot from attribute: {slot.name}")
        if remove_redundant_slot_usage:
            self.remove_redundant_slot_usage(schema)

    def merge_slot_usage(
        self,
        sv: SchemaView,
        cls: ClassDefinition,
        slot: SlotDefinition,
        overwrite=False,
    ):
        """
        Merge a SlotDefinition into the class_usage for a class

        If the class has no slot usage defined for that slot, then the class usage
        will be set to the input slot

        otherwise, each metaslot will be individually merged

        :param sv:
        :param cls:
        :param slot:
        :param overwrite: if True then overwrite conflicting values rather than raising error
        :return:
        """
        if slot.name not in cls.slot_usage:
            cls.slot_usage[slot.name] = slot
        else:
            su = cls.slot_usage[slot.name]
            for k, v in vars(slot).items():
                if v is not None and v != [] and v != {}:
                    curr_v = getattr(su, k, None)
                    if not overwrite and curr_v and curr_v != v:
                        raise ValueError(f"Conflict in {cls.name}.{slot.name}, attr {k} {v} != {curr_v}")
                    setattr(su, k, v)
            self._add_history(f"Merged slot usage: {slot.name}")

    def remove_redundant_slot_usage(self, schema: SchemaDefinition, class_name: ClassDefinitionName = None):
        """
        Remove parts of slot_usage that can be inferred

        :param schema:
        :param class_name:
        :return:
        """
        sv = SchemaView(schema)
        if class_name is None:
            for class_name in sv.all_classes():
                self.remove_redundant_slot_usage(schema, class_name)
        else:
            cls = sv.get_class(class_name)
            # test every key-value pair in the slot usage to determine which
            # slots within that are redundant
            slot_usage_keys = list(cls.slot_usage.keys())
            for slot_usage_key in slot_usage_keys:
                logger.debug(f"TESTING: {class_name}.{slot_usage_key}")
                slot_usage_value = cls.slot_usage[slot_usage_key]
                # perform a deletion test: what can be retrieved by inference
                del cls.slot_usage[slot_usage_key]
                sv.set_modified()
                try:
                    induced_slot = sv.induced_slot(slot_usage_key, class_name)
                except ValueError:
                    logger.warning(f"slot_usage with no slot: {slot_usage_key}")
                    continue
                # restore value
                cls.slot_usage[slot_usage_key] = slot_usage_value
                sv.set_modified()
                to_delete = []
                for metaslot_name, metaslot in vars(slot_usage_value).items():
                    if metaslot_name == "name":
                        # redundant with key
                        to_delete.append(metaslot_name)
                        continue
                    if metaslot_name in ["from_schema", "owner", "domain_of", "definition_uri"]:
                        # metamodel readonly slots are redundant by definition
                        to_delete.append(metaslot_name)
                        continue
                    v = getattr(slot_usage_value, metaslot_name, None)
                    if isinstance(v, bool) and not v:
                        # booleans with value False are inherently redundant
                        to_delete.append(metaslot_name)
                        continue
                    induced_v = getattr(induced_slot, metaslot_name, None)
                    if v is not None and v != [] and v != {} and v == induced_v:
                        logger.info(f"REDUNDANT: {class_name}.{slot_usage_key}[{metaslot_name}] = {v}")
                        to_delete.append(metaslot_name)
                for metaslot_name in to_delete:
                    del slot_usage_value[metaslot_name]
                    self._add_history(f"Removed redundant: {class_name}.slot_usage[{slot_usage_key}].[{metaslot_name}]")
            empty_keys = []
            for slot_usage_key, slot_usage_value in cls.slot_usage.items():
                metaslot_keys = list(json_dumper.to_dict(slot_usage_value).keys())
                if metaslot_keys == [] or metaslot_keys == ["name"]:
                    empty_keys.append(slot_usage_key)
            for k in empty_keys:
                del cls.slot_usage[k]

    @staticmethod
    def implicit_slots(schema: SchemaDefinition) -> dict[str, dict]:
        """
        Find slots that are implicit in the schema from slot_usage

        :param schema:
        :return:
        """
        sv = SchemaView(schema)
        implicit_slots1 = defaultdict(list)
        for cls in sv.all_classes().values():
            for slot in cls.slot_usage.values():
                slot_name = slot.name
                if slot_name not in sv.all_slots():
                    implicit_slots1[slot_name].append(json_dumper.to_dict(slot))
        new_slots = {}
        for slot_name, slot_list in implicit_slots1.items():
            all_keys = set()
            for slot in slot_list:
                all_keys.update(slot.keys())
            harmonized_slot = {}
            for k in all_keys:
                vals = []
                vals_strs = set()
                for slot in slot_list:
                    val = slot.get(k, None)
                    vals_strs.add(str(val))
                    vals.append(val)
                if len(vals_strs) == 1:
                    harmonized_slot[k] = vals.pop()
                elif len(vals_strs) > 1:
                    logger.info(f"Variable values in {slot_name}.{k}: {vals_strs}")
            new_slots[str(slot_name)] = harmonized_slot
        return new_slots

    def remove_unused_prefixes(self, schema: SchemaDefinition):
        raise NotImplementedError

    def _add_history(self, txt: str):
        if self.history is None:
            self.history = []
        self.history.append(txt)

    @staticmethod
    def fix_element_names(
        schema: SchemaDefinition,
        schema_dict: dict[str, Any] = None,
        rules: dict[str, Callable] = None,
        imports=False,
        preserve_original_using: Optional[str] = None,
    ) -> Union[YAMLRoot, dict]:
        """
        Changes element names to conform to naming conventions.

        :param schema: input schema
        :param schema_dict: if specified, the transformation will happen on this dictionary object
        :param rules: mappings between index slots and functions that normalize names
        :param imports: if True, all that imported modules are also fixed
        :param preserve_original_using: if specified, the original name will be preserved in this slot
        :return:
        """
        if rules is None:
            rules = {
                ClassDefinition.__name__: camelcase,
                TypeDefinition.__name__: camelcase,
                SlotDefinition.__name__: underscore,
                EnumDefinition.__name__: camelcase,
            }
        fixes = {}
        sv = SchemaView(schema)
        preserved = []
        for n, e in sv.all_elements(imports=imports).items():
            if e.from_schema == "https://w3id.org/linkml/types":
                continue
            typ = type(e).__name__
            if typ in rules:
                func = rules[typ]
                normalized = func(n)
                if normalized != n:
                    fixes[n] = normalized
                if preserve_original_using is not None:
                    preserved.append((typ, normalized, n))
                # if preserve_original_using is not None:
                #    setattr(e, preserve_original_using, n)
                #    print(f"SETTING {typ} {e.name}.{preserve_original_using} = {n}")
        if schema_dict is not None:
            schema = schema_dict
        schema = yaml_rewrite(schema, fixes)
        for typ, normalized, original in preserved:
            pathmap = {
                ClassDefinition.__name__: "classes",
                TypeDefinition.__name__: "types",
                SlotDefinition.__name__: "slots",
                EnumDefinition.__name__: "enums",
            }
            if isinstance(schema, dict):
                path = schema[pathmap[typ]]
                if normalized not in path:
                    logger.warning(f"Cannot find {typ} {normalized} in {pathmap[typ]}")
                    continue
                e = path[normalized]
                if preserve_original_using not in e:
                    path[normalized][preserve_original_using] = original
            else:
                path = getattr(schema, pathmap[typ])
                e = path[normalized]
                if not getattr(e, preserve_original_using, None):
                    setattr(e, preserve_original_using, original)
        return schema


@click.group(name="fix")
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet")
def main(verbose: int, quiet: bool):
    """Apply schema-fixer commands."""
    if verbose >= 2:
        logger.setLevel(level=logging.DEBUG)
    elif verbose == 1:
        logger.setLevel(level=logging.INFO)
    else:
        logger.setLevel(level=logging.WARNING)
    if quiet:
        logger.setLevel(level=logging.ERROR)


@main.command()
@click.argument("input_schema")
@click.option(
    "--imports/--no-imports",
    default=False,
    show_default=True,
    help="Apply fix to referenced elements from modules",
)
@click.option(
    "--preserve-original-using",
    "-P",
    default=None,
    show_default=True,
    help="If specified, original name will be preserved in this slot (e.g. title)",
)
def fix_names(input_schema, **kwargs):
    """Fix element names to conform to naming conventions"""
    with open(input_schema) as f:
        schema_dict = yaml.safe_load(f)
    sv = SchemaView(input_schema)
    fixer = SchemaFixer()
    schema = fixer.fix_element_names(sv.schema, schema_dict, **kwargs)
    print(yaml.dump(schema, sort_keys=False))


@main.command()
@click.argument("input_schema")
@click.option(
    "--imports/--no-imports",
    default=False,
    show_default=True,
    help="Apply fix to referenced elements from modules",
)
def implicit_slots(input_schema, **kwargs):
    """Find implicit slots in schema"""
    with open(input_schema) as f:
        yaml.safe_load(f)
    sv = SchemaView(input_schema)
    fixer = SchemaFixer()
    slots = fixer.implicit_slots(sv.schema)
    for slot in slots.values():
        del slot["name"]
    print(yaml.dump({"slots": slots}, sort_keys=False))


if __name__ == "__main__":
    main()



import polars as pl

from .model_transform import ModelTransform


class SimpleDictModelTransform(ModelTransform):
    """This class assists in converting a LinkML 'simple dict' inline column
    into a form that is better for representing in a PolaRS dataframe and
    validating with a Pandera model.
    """

    def __init__(self, polars_schema, id_col, other_col):
        self.polars_schema = polars_schema
        """A polars schema representing a simple dict column"""

        self.id_col = id_col
        """The ID column in the sense of a LinkML inline simple dict"""

        self.other_col = other_col
        """The 'other' column in the sense of a LinkML inline simple dict"""

        self.id_col_type = None
        self.other_col_type = None
        self.polars_struct = self._build_polars_struct()
        """A pl.Struct representing the schema of the other range."""

    def _build_polars_struct_simple(self):
        """Handles the two column (id, other) form of the simple dict"""
        self.id_col_type = self.polars_schema.columns[self.id_col].dtype.type
        self.other_col_type = self.polars_schema.columns[self.other_col].dtype.type

        return pl.Struct({self.id_col: self.id_col_type, self.other_col: self.other_col_type})

    def _build_polars_struct_complex(self):
        """Handles the non-two-column simple dict cases."""
        struct_items = {}
        for k, v in self.polars_schema.columns.items():
            if v.dtype.type == pl.Object:
                v.dtype.type = pl.Struct
            else:
                struct_items[k] = v.dtype.type
        return pl.Struct(struct_items)

    def _build_polars_struct(self):
        if len(self.polars_schema.columns.keys()) == 2:
            return self._build_polars_struct_simple()
        else:
            return self._build_polars_struct_complex()

    def transform(self, linkml_simple_dict):
        """Converts a simple dict nested column to a list of dicts.
        { 'A': 1, 'B': 2, ... } -> [{'id': 'other': 1}, {'id': 'B', 'other': 2}, ...]
        """
        return self._simple_dict_to_list_of_structs(linkml_simple_dict)

    def _simple_dict_to_list_of_structs(self, linkml_simple_dict):
        """Converts a simple dict nested column to a list of dicts.
        { 'A': 1, 'B': 2, ... } -> [{'id': 'other': 1}, {'id': 'B', 'other': 2}, ...]

        An inefficient conversion (relative to native PolaRS operations)
        from a simple dict form to a dataframe struct column.

        e : dict
            e is a single row entry in a dataframe column (one cell), which itself is a dict.
            The value entries of e may also be dicts.
        """
        arr = []
        for id_value, range_value in linkml_simple_dict.items():
            if isinstance(range_value, dict) and (set(range_value.keys()) <= set(self.polars_schema.columns.keys())):
                range_dict = range_value
                range_dict[self.id_col] = id_value
                for column_key in self.polars_schema.columns.keys():
                    if column_key not in range_dict:
                        range_dict[column_key] = None
            else:
                range_dict = {self.id_col: id_value, self.other_col: range_value}
            arr.append(range_dict)

        return arr

    def list_dtype(self):
        return pl.List(self.polars_struct)

    def explode_unnest_dataframe(self, df, column_name):
        """Explode and unnest for simple dict."""
        return df.lazy().explode(column_name).unnest(column_name).collect()



from __future__ import annotations

import logging
import os
from collections import defaultdict
from dataclasses import dataclass
from types import ModuleType

import click
from jinja2 import Template
from linkml_runtime.linkml_model import Annotation, ClassDefinition, ClassDefinitionName, SchemaDefinition
from linkml_runtime.utils.compile_python import compile_python
from linkml_runtime.utils.formatutils import camelcase, underscore
from linkml_runtime.utils.schemaview import SchemaView
from sqlalchemy import Enum

from linkml._version import __version__
from linkml.generators.pydanticgen import PydanticGenerator
from linkml.generators.pythongen import PythonGenerator
from linkml.generators.sqlalchemy.sqlalchemy_declarative_template import sqlalchemy_declarative_template_str
from linkml.generators.sqlalchemy.sqlalchemy_imperative_template import sqlalchemy_imperative_template_str
from linkml.generators.sqltablegen import SQLTableGenerator
from linkml.transformers.relmodel_transformer import ForeignKeyPolicy, RelationalModelTransformer
from linkml.utils.generator import Generator, shared_arguments

logger = logging.getLogger(__name__)


class TemplateEnum(Enum):
    DECLARATIVE = "declarative"
    IMPERATIVE = "imperative"


@dataclass
class SQLAlchemyGenerator(Generator):
    """
    Generates SQL Alchemy classes

    See also: :class:`~linkml.generators.sqltablegen.SQLTableGenerator`
    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.1"
    valid_formats = ["sqla"]
    file_extension = "py"
    uses_schemaloader = False
    template: TemplateEnum | None = None

    # ObjectVars
    original_schema: SchemaDefinition | str = None

    def __post_init__(self) -> None:
        self.original_schema = self.schema
        self.schemaview = SchemaView(self.schema)
        super().__post_init__()

    def generate_sqla(
        self,
        model_path: str | None = None,
        no_model_import: bool = False,
        template: TemplateEnum | None = None,
        foreign_key_policy: ForeignKeyPolicy | None = None,
        **kwargs,
    ) -> str:
        template = template or self.template or TemplateEnum.IMPERATIVE
        sqltr = RelationalModelTransformer(self.schemaview)
        if foreign_key_policy:
            sqltr.foreign_key_policy = foreign_key_policy
        tgen = SQLTableGenerator(self.schemaview.schema)
        tr_result = sqltr.transform(**kwargs)
        tr_schema = tr_result.schema
        for c in tr_schema.classes.values():
            for a in c.attributes.values():
                sql_range = tgen.get_sql_range(a, tr_schema)
                sql_type = sql_range.__repr__()
                ann = Annotation("sql_type", sql_type)
                a.annotations[ann.tag] = ann
        if template == TemplateEnum.IMPERATIVE:
            template_str = sqlalchemy_imperative_template_str
        elif template == TemplateEnum.DECLARATIVE:
            template_str = sqlalchemy_declarative_template_str
        else:
            raise Exception(f"Unknown template type: {template}")
        template_obj = Template(template_str)
        if model_path is None:
            model_path = self.schema.name
        logger.info(f"Package for dataclasses ==  {model_path}")
        backrefs = defaultdict(list)
        for m in tr_result.mappings:
            backrefs[m.source_class].append(m)
        self.add_safe_aliases(tr_schema)
        tr_sv = SchemaView(tr_schema)
        rel_schema_classes_ordered = [tr_sv.get_class(cn, strict=True) for cn in self.order_classes_by_hierarchy(tr_sv)]
        rel_schema_classes_ordered = [c for c in rel_schema_classes_ordered if not self.skip(c)]
        for c in rel_schema_classes_ordered:
            # For SQLA there needs to be a primary key for each class;
            # autogenerate this as a compound key if none declared
            has_pk = any(a for a in c.attributes.values() if "primary_key" in a.annotations)
            if not has_pk:
                for a in c.attributes.values():
                    ann = Annotation("primary_key", "true")
                    a.annotations[ann.tag] = ann
        code = template_obj.render(
            model_path=model_path,
            mappings=tr_result.mappings,
            backrefs=backrefs,
            classname=camelcase,
            no_model_import=no_model_import,
            is_join_table=lambda c: any(tag for tag in c.annotations.keys() if tag == "linkml:derived_from"),
            classes=rel_schema_classes_ordered,
        )
        logger.debug(f"# Generated code:\n{code}")
        return code

    def serialize(self, **kwargs) -> str:
        return self.generate_sqla(**kwargs)

    def compile_sqla(
        self,
        compile_python_dataclasses=False,
        pydantic=False,
        model_path=None,
        template: TemplateEnum = TemplateEnum.IMPERATIVE,
        **kwargs,
    ) -> ModuleType:
        """
        Generates and compiles SQL Alchemy bindings

        - If template is DECLARATIVE, then a single python module with classes is generated
        - If template is IMPERATIVE, only mappings are generated
             - if compile_python_dataclasses is True then a standard datamodel is generated

        :param compile_python_dataclasses: (default False)
        :param pydantic:
        :param model_path:
        :param template:
        :param kwargs:
        :return:
        """

        if model_path is None:
            model_path = self.schema.name

        if template == TemplateEnum.DECLARATIVE:
            sqla_code = self.generate_sqla(model_path=None, no_model_import=True, template=template, **kwargs)
            return compile_python(sqla_code, package_path=model_path)
        elif compile_python_dataclasses:
            # concatenate the python dataclasses with the sqla code
            if pydantic:
                # mixin inheritance doesn't get along with SQLAlchemy's imperative (aka classical) mapping
                pygen = PydanticGenerator(self.original_schema, extra_fields="allow", gen_mixin_inheritance=False)
            else:
                pygen = PythonGenerator(self.original_schema)
            dc_code = pygen.serialize()
            sqla_code = self.generate_sqla(model_path=None, no_model_import=True, **kwargs)
            return compile_python(f"{dc_code}\n{sqla_code}", package_path=model_path)
        else:
            code = self.generate_sqla(model_path=model_path, **kwargs)
            return compile_python(code, package_path=model_path)

    @staticmethod
    def add_safe_aliases(schema: SchemaDefinition) -> None:
        for c in schema.classes.values():
            for a in c.attributes.values():
                a.alias = underscore(a.name)

    @staticmethod
    def skip(cls: ClassDefinition) -> bool:
        is_skip = len(cls.attributes) == 0
        if is_skip:
            logger.error(f"SKIPPING: {cls.name}")
        return is_skip

    # TODO: move this
    @staticmethod
    def order_classes_by_hierarchy(sv: SchemaView) -> list[ClassDefinitionName]:
        olist = sv.class_roots()
        unprocessed = [cn for cn in sv.all_classes() if cn not in olist]
        while len(unprocessed) > 0:
            ext_list = [cn for cn in unprocessed if not any(p for p in sv.class_parents(cn) if p not in olist)]
            if len(ext_list) == 0:
                raise ValueError(f"Cycle in hierarchy, cannot process: {unprocessed}")
            olist += ext_list
            unprocessed = [cn for cn in unprocessed if cn not in olist]
        return olist


@shared_arguments(SQLAlchemyGenerator)
@click.option(
    "--declarative/--no-declarative",
    default=True,
    show_default=True,
    help="Generate SQL Alchemy declarative vs imperative",
)
@click.option(
    "--generate-classes/--no-generate-classes",
    default=False,
    show_default=True,
    help="If True, generate Python datamodel (imperative mode only)",
)
@click.option(
    "--pydantic/--no-pydantic",
    default=False,
    show_default=True,
    help="If True, generate Pydantic classes (imperative mode only)",
)
@click.option(
    "--use-foreign-keys/--no-use-foreign-keys",
    default=True,
    show_default=True,
    help="Emit FK declarations",
)
@click.version_option(__version__, "-V", "--version")
@click.command(name="sqla")
def cli(yamlfile, declarative, generate_classes, pydantic, use_foreign_keys=True, **args):
    """Generate SQL DDL representation"""
    if pydantic:
        pygen = PydanticGenerator(yamlfile)
        print(pygen.serialize())
    gen = SQLAlchemyGenerator(yamlfile, **args)
    if declarative:
        t = TemplateEnum.DECLARATIVE
    else:
        t = TemplateEnum.IMPERATIVE
    if use_foreign_keys:
        foreign_key_policy = None  # default
    else:
        foreign_key_policy = ForeignKeyPolicy.NO_FOREIGN_KEYS
    print(gen.generate_sqla(template=t, foreign_key_policy=foreign_key_policy))
    if generate_classes:
        raise NotImplementedError("generate classes not implemented")


if __name__ == "__main__":
    cli()



import sys
from importlib.util import find_spec
from keyword import iskeyword
from typing import Any, ClassVar, Literal, Optional, Union, get_args

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

from jinja2 import Environment, PackageLoader
from pydantic import BaseModel, Field, field_validator, model_validator
from pydantic.version import VERSION as PYDANTIC_VERSION

from linkml.generators.common.template import (
    ConditionalImport as ConditionalImport_,
)
from linkml.generators.common.template import (
    Import as Import_,
)
from linkml.generators.common.template import (
    Imports as Imports_,
)
from linkml.generators.common.template import (
    ObjectImport,  # noqa: F401
    TemplateModel,
)
from linkml.utils.deprecation import deprecation_warning

try:
    if find_spec("black") is not None:
        from linkml.generators.pydanticgen.black import format_black
    else:
        # no warning, having black is optional, we only warn when someone tries to import it explicitly
        format_black = None
except ImportError:
    # we can also get an import error from find_spec during testing because that's how we mock not having it installed
    format_black = None

if int(PYDANTIC_VERSION[0]) >= 2:
    from pydantic import computed_field
else:
    deprecation_warning("pydantic-v1")

    def computed_field(f):
        """No-op decorator to allow this module to not break imports until 1.9.0"""
        return f


IMPORT_GROUPS = Literal["future", "stdlib", "thirdparty", "local", "conditional"]
"""
See :attr:`.Import.group` and :attr:`.Imports.sort`

Order of this literal is used in sort and therefore not arbitrary.
"""


class PydanticTemplateModel(TemplateModel):
    """
    Metaclass to render pydantic models with jinja templates.

    Each subclass needs to declare a :class:`typing.ClassVar` for a
    jinja template within the `templates` directory.

    Templates are written expecting each of the other TemplateModels
    to already be rendered to strings - ie. rather than the ``class.py.jinja``
    template receiving a full :class:`.PydanticAttribute` object or dictionary,
    it receives it having already been rendered to a string. See the :meth:`.render` method.

    .. admonition:: Black Formatting

        Template models will try to use ``black`` to format results when it is available in the
        environment when render is called with ``black = True`` . If it isn't, then the string is
        returned without any formatting beyond the template.
        This is mostly important for complex annotations like those produced for arrays,
        as otherwise the templates are acceptable looking.

        To install linkml with black, use the extra ``black`` dependency.

        e.g. with pip::

            pip install linkml[black]

        or with poetry::

            poetry install -E black

    """

    template: ClassVar[str]
    _environment: ClassVar[Environment] = Environment(
        loader=PackageLoader("linkml.generators.pydanticgen", "templates"), trim_blocks=True, lstrip_blocks=True
    )

    meta_exclude: ClassVar[list[str]] = None

    def render(self, environment: Optional[Environment] = None, black: bool = False) -> str:
        """
        Recursively render a template model to a string.

        For each field in the model, recurse through, rendering each :class:`.PydanticTemplateModel`
        using the template set in :attr:`.PydanticTemplateModel.template` , but preserving the structure
        of lists and dictionaries. Regular :class:`.BaseModel` s are rendered to dictionaries.
        Any other value is passed through unchanged.

        Args:
            environment (:class:`jinja2.Environment`): Template environment - see :meth:`.environment`
            black (bool): if ``True`` , format template with black. (default False)
        """
        if environment is None:
            environment = self.environment()

        rendered = super().render(environment=environment)

        if format_black is not None and black:
            try:
                return format_black(rendered)
            except Exception:
                # TODO: it would nice to have a standard logging module here ;)
                return rendered
        elif black and format_black is None:
            raise ValueError("black formatting was requested, but black is not installed in this environment")
        else:
            return rendered


class EnumValue(BaseModel):
    """
    A single value within an :class:`.Enum`
    """

    label: str
    alias: Optional[str] = None
    value: str
    description: Optional[str] = None

    @model_validator(mode="after")
    def alias_python_keywords(self) -> Self:
        """Mask Python keywords used for `label` by appending `_`"""
        if iskeyword(self.label):
            if self.alias is None:
                self.alias = self.label
            self.label = self.label + "_"
        return self


class PydanticEnum(PydanticTemplateModel):
    """
    Model used to render a :class:`enum.Enum`
    """

    template: ClassVar[str] = "enum.py.jinja"

    name: str
    description: Optional[str] = None
    values: dict[str, EnumValue] = Field(default_factory=dict)


class PydanticBaseModel(PydanticTemplateModel):
    """
    Parameterization of the base model that generated pydantic classes inherit from
    """

    template: ClassVar[str] = "base_model.py.jinja"

    default_name: ClassVar[str] = "ConfiguredBaseModel"
    name: str = Field(default_factory=lambda: PydanticBaseModel.default_name)
    extra_fields: Literal["allow", "forbid", "ignore"] = "forbid"
    """
    Sets the ``extra`` model for pydantic models
    """
    fields: Optional[list[str]] = None
    """
    Extra fields that are typically injected into the base model via
    :attr:`~linkml.generators.pydanticgen.PydanticGenerator.injected_fields`
    """
    strict: bool = False
    """
    Enable strict mode in the base model.

    .. note::

        Pydantic 2 only! Pydantic 1 only has strict types, not strict mode. See: https://github.com/linkml/linkml/issues/1955

    References:
        https://docs.pydantic.dev/latest/concepts/strict_mode
    """


class PydanticAttribute(PydanticTemplateModel):
    """
    Reduced version of SlotDefinition that carries all and only the information
    needed by the template
    """

    template: ClassVar[str] = "attribute.py.jinja"
    meta_exclude: ClassVar[list[str]] = ["from_schema", "owner", "range", "inlined", "inlined_as_list"]

    name: str
    alias: Optional[str] = None
    required: bool = False
    identifier: bool = False
    key: bool = False
    predefined: Optional[str] = None
    """Fixed string to use in body of field"""
    range: Optional[str] = None
    """Type annotation used for model field"""
    title: Optional[str] = None
    description: Optional[str] = None
    equals_number: Optional[Union[int, float]] = None
    minimum_value: Optional[Union[int, float]] = None
    maximum_value: Optional[Union[int, float]] = None
    exact_cardinality: Optional[int] = None
    minimum_cardinality: Optional[int] = None
    maximum_cardinality: Optional[int] = None
    multivalued: Optional[bool] = None
    pattern: Optional[str] = None
    meta: Optional[dict[str, Any]] = None
    """
    Metadata for the slot to be included in a Field annotation
    """

    @computed_field
    def field(self) -> str:
        """Computed value to use inside of the generated Field"""
        if self.predefined:
            return self.predefined
        elif self.required or self.identifier or self.key:
            return "..."
        else:
            if self.range and self.range.startswith("Optional[list"):
                return "[]"
            return "None"

    @model_validator(mode="after")
    def alias_python_keywords(self) -> Self:
        """Mask Python keywords used for `name` by appending `_`"""
        if iskeyword(self.name):
            if self.alias is None:
                self.alias = self.name
            self.name = self.name + "_"
        return self


class PydanticValidator(PydanticAttribute):
    """
    Trivial subclass of :class:`.PydanticAttribute` that uses the ``validator.py.jinja`` template instead
    """

    template: ClassVar[str] = "validator.py.jinja"


class PydanticClass(PydanticTemplateModel):
    """
    Reduced version of ClassDefinition that carries all and only the information
    needed by the template.

    On instantiation and rendering, will create any additional :attr:`.validators`
    that are implied by the given :attr:`.attributes`. Currently the only kind of
    slot-level validators that are created are for those slots that have a ``pattern``
    property.
    """

    template: ClassVar[str] = "class.py.jinja"
    meta_exclude: ClassVar[list[str]] = ["slots", "is_a"]

    name: str
    bases: Union[list[str], str] = PydanticBaseModel.default_name
    description: Optional[str] = None
    attributes: Optional[dict[str, PydanticAttribute]] = None
    meta: Optional[dict[str, Any]] = None
    """
    Metadata for the class to be included in a linkml_meta class attribute
    """

    def _validators(self) -> Optional[dict[str, PydanticValidator]]:
        if self.attributes is None:
            return None

        return {k: PydanticValidator(**v.model_dump()) for k, v in self.attributes.items() if v.pattern is not None}

    @computed_field
    def validators(self) -> Optional[dict[str, PydanticValidator]]:
        return self._validators()

    @computed_field
    def slots(self) -> Optional[dict[str, PydanticAttribute]]:
        """alias of attributes"""
        return self.attributes


class Import(Import_, PydanticTemplateModel):
    """
    A python module, or module and classes to be imported.

    Examples:

        Module import:

        .. code-block:: python

            >>> Import(module='sys').render()
            import sys
            >>> Import(module='numpy', alias='np').render()
            import numpy as np

        Class import:

        .. code-block:: python

            >>> Import(module='pathlib', objects=[
            >>>     ObjectImport(name="Path"),
            >>>     ObjectImport(name="PurePath", alias="RenamedPurePath")
            >>> ]).render()
            from pathlib import (
                Path,
                PurePath as RenamedPurePath
            )

    """

    template: ClassVar[str] = "imports.py.jinja"

    @computed_field
    def group(self) -> IMPORT_GROUPS:
        """
        Import group used when sorting

        * ``future`` - from `__future__` import...
        * ``stdlib`` - ... the standard library
        * ``thirdparty`` - other dependencies not in the standard library
        * ``local`` - relative imports (eg. from split generation)
        * ``conditional`` - a :class:`.ConditionalImport`
        """
        if self.module == "__future__":
            return "future"
        elif sys.version_info.minor >= 10 and self.module in sys.stdlib_module_names:
            return "stdlib"
        elif sys.version_info.minor < 10 and self.module in _some_stdlib_module_names:
            return "stdlib"
        elif self.module.startswith("."):
            return "local"
        else:
            return "thirdparty"


class ConditionalImport(ConditionalImport_, PydanticTemplateModel):
    """
    Import that depends on some condition in the environment, common when
    using backported features or straddling dependency versions.

    Make sure that everything that is needed to evaluate the condition is imported
    before this is added to the injected imports!

    Examples:

        conditionally import Literal from ``typing_extensions`` if on python <= 3.8

        .. code-block:: python
            :force:

            imports = (Imports() +
                 Import(module='sys') +
                 ConditionalImport(
                 module="typing",
                 objects=[ObjectImport(name="Literal")],
                 condition="sys.version_info >= (3, 8)",
                 alternative=Import(
                     module="typing_extensions",
                     objects=[ObjectImport(name="Literal")]
                 )
             )

        Renders to:

        .. code-block:: python
            :force:

            import sys
            if sys.version_info >= (3, 8):
                from typing import Literal
            else:
                from typing_extensions import Literal

    """

    template: ClassVar[str] = "conditional_import.py.jinja"

    @computed_field
    def group(self) -> Literal["conditional"]:
        return "conditional"


class Imports(Imports_, PydanticTemplateModel):
    """
    Container class for imports that can handle merging!

    See :class:`.Import` and :class:`.ConditionalImport` for examples of declaring individual imports

    Useful for generation, because each build stage will potentially generate
    overlapping imports. This ensures that we can keep a collection of imports
    without having many duplicates.

    Defines methods for adding, iterating, and indexing from within the :attr:`Imports.imports` list.

    Examples:

        .. code-block:: python
            :force:

            imports = (Imports() +
                Import(module="sys") +
                Import(module="pathlib", objects=[ObjectImport(name="Path")]) +
                Import(module="sys")
            )

        Renders to:

        .. code-block:: python

            from pathlib import Path
            import sys

    """

    template: ClassVar[str] = "imports.py.jinja"

    imports: list[Union[Import, ConditionalImport]] = Field(default_factory=list)
    group_order: tuple[str, ...] = get_args(IMPORT_GROUPS)
    """Order in which to sort imports by their :attr:`.Import.group`"""

    @computed_field
    def import_groups(self) -> list[IMPORT_GROUPS]:
        """
        List of what group each import belongs to
        """
        return [i.group for i in self.imports]

    def sort(self) -> None:
        """
        Sort imports recursively, mimicking isort:

        * First by :attr:`.Import.group` according to :attr:`.Imports.group_order`
        * Then by whether the :class:`.Import` has any objects
          (``import module`` comes before ``from module import name``)
        * Then alphabetically by module name
        """

        def _sort_key(i: Import) -> tuple[int, int, str]:
            return (self.group_order.index(i.group), int(i.objects is not None), i.module)

        imports = sorted(self.imports, key=_sort_key)
        for i in imports:
            i.sort()
        self.imports = imports


class PydanticModule(PydanticTemplateModel):
    """
    Top-level container model for generating a pydantic module :)
    """

    template: ClassVar[str] = "module.py.jinja"
    meta_exclude: ClassVar[str] = ["slots"]

    metamodel_version: Optional[str] = None
    version: Optional[str] = None
    base_model: PydanticBaseModel = PydanticBaseModel()
    injected_classes: Optional[list[str]] = None
    python_imports: Union[Imports, list[Union[Import, ConditionalImport]]] = Imports()
    enums: dict[str, PydanticEnum] = Field(default_factory=dict)
    classes: dict[str, PydanticClass] = Field(default_factory=dict)
    meta: Optional[dict[str, Any]] = None
    """
    Metadata for the schema to be included in a linkml_meta module-level instance of LinkMLMeta
    """

    @field_validator("python_imports", mode="after")
    @classmethod
    def cast_imports(cls, imports: Union[Imports, list[Union[Import, ConditionalImport]]]) -> Imports:
        if isinstance(imports, list):
            imports = Imports(imports=imports)
        return imports

    @computed_field
    def class_names(self) -> list[str]:
        return [c.name for c in self.classes.values()]


_some_stdlib_module_names = {
    "copy",
    "datetime",
    "decimal",
    "enum",
    "inspect",
    "os",
    "re",
    "sys",
    "typing",
    "dataclasses",
}
"""
sys.stdlib_module_names is only present in 3.10 and later
so we make a cheap copy of the stdlib modules that we commonly use here,
but this should be removed whenever support for 3.9 is dropped.
"""
